#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""Tools for building and applying constraints and penalties.
"""

__all__ = ['with_penalty','with_constraint','as_penalty','as_constraint',
           'with_mean','with_variance','with_std','with_spread','normalized',
           'issolution','solve','discrete','integers','rounded','precision',
           'near_integers','unique','has_unique','impose_unique','bounded',
           'impose_bounds','impose_as','impose_at','impose_measure',
           'impose_position','impose_weight','and_','or_','not_','vectorize',
           'boundsconstrain','monotonic','sorting','periodic']

from mystic.math.measures import *
from mystic.math import almostEqual

def vectorize(constraint, axis=1):
    """vectorize a 1D constraint function ``x' = c(x)`` for 2D input and output

    Args:
        constraint (function): a function ``c``, where ``x' = c(x)`` with ``x``
          and ``x'`` lists of the same length
        axis (int, default=1): index of the axis of ``x`` to apply constraints;
          must be either 0 or 1

    Returns:
        a function ``k``, where ``x' = k(x)`` with ``x`` and ``x'`` 2D arrays
          of the same shape

    Notes:
        - Produces a ``transform`` function that is of the form required by
          ``sklearn.preprocessing.FunctionTransformer(func=transform)``, and
          takes a ``numpy`` array of shape ``(samples, features)`` as input.

    Examples:
        >>> from mystic.constraints import (impose_bounds, integers,
        ...                                 with_mean, and_)
        >>> cons = and_(impose_bounds([(0,5),(7,10)])(lambda x:x),
        ...             integers()(lambda x:x), with_mean(6.0)(lambda x:x))
        >>> import numpy as np
        >>> data = np.random.randn(6,4)
        >>> c = vectorize(cons, axis=0)
        >>> c(data)
        array([[ 3,  9,  3, 10],
               [ 7,  8,  7,  4],
               [ 9,  3,  7,  7],
               [ 7,  3,  8,  7],
               [ 3,  5,  4,  4],
               [ 7,  8,  7,  4]])
        >>> _.mean(axis=0)
        array([6., 6., 6., 6.])
        >>> c = vectorize(cons, axis=1)
        >>> c(data)
        array([[ 3,  3,  9,  9],
               [ 8,  3,  4,  9],
               [ 5, 10,  4,  5],
               [ 7,  8,  7,  2],
               [ 2,  4,  8, 10],
               [ 7, 10,  5,  2]])
        >>> _.mean(axis=1)
        array([6., 6., 6., 6., 6., 6.])
        >>> k = FunctionTransformer(func=c)
        >>> k.fit(data).transform(data).mean(axis=1)
        array([6., 6., 6., 6., 6., 6.])
    """
    if not axis in (0,1):
        msg = "axis = %s, must be either 0 or 1" % axis
        raise ValueError(msg)
    _doc = """mystic kernel transform `x' = k(x)` with x a 2D numpy array.

        """
    if axis:
        def transform(x, *args, **kwds):
            import numpy as np
            c = lambda x,*args,**kwds: np.array([constraint(xi, *args, **kwds) for xi in x])
            return c(x, *args, **kwds)
    else:
        def transform(x, *args, **kwds):
            import numpy as np
            c = lambda x,*args,**kwds: np.array([constraint(xi, *args, **kwds) for xi in x.T]).T
            return c(x, *args, **kwds)
    doc = constraint.__doc__
    transform.__doc__ = _doc + doc if doc else _doc
    return transform


def with_penalty(ptype, *args, **kwds):
    """convert a condition to a penalty function of the chosen type

condition f(x) is satisfied when f(x) == 0.0 for equality constraints
and f(x) <= 0.0 for inequality constraints. ptype is a mystic.penalty type.

Examples:
    >>> @with_penalty(quadratic_equality, kwds={'target':5.0})
    ... def penalty(x, target):
    ...   return mean(x) - target
    >>> 
    >>> penalty([1,2,3,4,5])
    400.0
    >>> penalty([3,4,5,6,7])
    7.8886090522101181e-29
    """
    def dec(condition):
      @ptype(condition, *args, **kwds)
      def penalty(x):
          return 0.0
      penalty.func = condition
      penalty.ptype = ptype.__name__ 
      return penalty
    return dec


#XXX: better: wrap_constraint( constraint, func, ctype=inner) ?
def with_constraint(ctype, *args, **kwds): #XXX: is this *at all* useful?
    """convert a set transformation to a constraints solver of the chosen type

transformation f(x) is a mapping between x and x', where x' = f(x).
ctype is a mystic.coupler type [inner, outer, inner_proxy, outer_proxy].

Examples:
    >>> @with_constraint(inner, kwds={'target':5.0})
    ... def constraint(x, target):
    ...   return impose_mean(target, x)
    ... 
    >>> x = constraint([1,2,3,4,5])
    >>> print(x)
    [3.0, 4.0, 5.0, 6.0, 7.0]
    >>> mean(x)
    5.0
    """
    def dec(condition):
      @ctype(condition, *args, **kwds)
      def constraint(x):
          return x
      return constraint
    return dec


# XXX: provide the below from measures, which from bounded?, and ...generic?
'''
mean(samples, weights=None)
variance(samples, weights=None)
spread(samples)
expectation(f, samples, weights=None, tol=0.0)
# impose_mean(m, samples, weights=None)
# impose_variance(v, samples, weights=None)
# impose_spread(r, samples, weights=None)
impose_expectation(m, f, npts, bounds=None, weights=None, tol=None, **kwds)
impose_weight_norm(samples, weights, mass=1.0)
# normalize(weights, mass=1.0, zsum=False, zmass=1.0)
'''

def with_mean(target):
    """bind a mean constraint to a given constraints function.

Inputs:
    target -- the target mean

A constraints function takes an iterable x as input, returning a modified x.
This function is an "outer" coupling of "impose_mean" onto another constraints
function c(x), such that:  x' = impose_mean(target, c(x)).

Examples:
    >>> @with_mean(5.0)
    ... def constraint(x):
    ...   x[-1] = x[0]
    ...   return x
    ... 
    >>> x = constraint([1,2,3,4])
    >>> print(x)
    [4.25, 5.25, 6.25, 4.25]
    >>> mean(x)
    5.0
    """
    def decorate(constraints):
        def factory(x, *args, **kwds):
            # apply decorated constraints function
            x = constraints(x, *args, **kwds)
            # constrain x such that mean(x) == target
            if not almostEqual(mean(x), target):
                x = impose_mean(target, x)#, weights=weights)
            return x
        return factory
    return decorate


def with_variance(target):
    """bind a variance constraint to a given constraints function.

Inputs:
    target -- the target variance

A constraints function takes an iterable x as input, returning a modified x.
This function is an "outer" coupling of "impose_variance" onto another
constraints function c(x), such that:  x' = impose_variance(target, c(x)).

Examples:
    >>> @with_variance(1.0)
    ... def constraint(x):
    ...   x[-1] = x[0]
    ...   return x
    ... 
    >>> x = constraint([1,2,3])
    >>> print(x)
    [0.6262265521467858, 2.747546895706428, 0.6262265521467858]
    >>> variance(x)
    0.99999999999999956
    """
    def decorate(constraints):
        def factory(x, *args, **kwds):
            # apply decorated constraints function
            x = constraints(x, *args, **kwds)
            # constrain x such that variance(x) == target
            if not almostEqual(variance(x), target):
                x = impose_variance(target, x)#, weights=weights)
            return x
        return factory
    return decorate


def with_std(target):
    """bind a standard deviation constraint to a given constraints function.

Inputs:
    target -- the target standard deviation

A constraints function takes an iterable x as input, returning a modified x.
This function is an "outer" coupling of "impose_std" onto another
constraints function c(x), such that:  x' = impose_std(target, c(x)).

Examples:
    >>> @with_std(1.0)
    ... def constraint(x):
    ...   x[-1] = x[0]
    ...   return x
    ... 
    >>> x = constraint([1,2,3])
    >>> print(x)
    [0.6262265521467858, 2.747546895706428, 0.6262265521467858]
    >>> std(x)
    0.99999999999999956
    """
    return with_variance(target**2)


def with_spread(target):
    """bind a range constraint to a given constraints function.

Inputs:
    target -- the target range

A constraints function takes an iterable x as input, returning a modified x.
This function is an "outer" coupling of "impose_spread" onto another constraints
function c(x), such that:  x' = impose_spread(target, c(x)).

Examples:
    >>> @with_spread(10.0)
    ... def constraint(x):
    ...   return [i**2 for i in x]
    ... 
    >>> x = constraint([1,2,3,4])
    >>> print(x)
    [3.1666666666666665, 5.1666666666666661, 8.5, 13.166666666666666]
    >>> spread(x)
    10.0
    """
    def decorate(constraints):
        def factory(x, *args, **kwds):
            # apply decorated constraints function
            x = constraints(x, *args, **kwds)
            # constrain x such that spread(x) == target
            if not almostEqual(spread(x), target):
                x = impose_spread(target, x)#, weights=weights)
            return x
        return factory
    return decorate


from numpy import sum
def normalized(mass=1.0): #XXX: order matters when multiple decorators
    """bind a normalization constraint to a given constraints function.

Inputs:
    mass -- the target sum of normalized weights

A constraints function takes an iterable x as input, returning a modified x.
This function is an "outer" coupling of "normalize" onto another constraints
function c(x), such that:  x' = normalize(c(x), mass).

Examples:
    >>> @normalized()
    ... def constraint(x):
    ...   return x
    ... 
    >>> constraint([1,2,3])
    [0.16666666666666666, 0.33333333333333331, 0.5]
    """
    def decorate(constraints):
        def factory(x, *args, **kwds):
            # apply decorated constraints function
            x = constraints(x, *args, **kwds)
            # constrain x such that sum(x) == mass
            if not almostEqual(sum(x), mass):
                x = normalize(x, mass=mass)
            return x
        return factory
    return decorate


def issolution(constraints, guess, tol=1e-3):
    """Returns whether the guess is a solution to the constraints

Input:
    constraints -- a constraints solver function or a penalty function
    guess -- list of parameter values proposed to solve the constraints
    tol -- residual error magnitude for which constraints are considered solved

Examples:
    >>> @normalized()
    ... def constraint(x):
    ...   return x
    ... 
    >>> constraint([.5,.5])
    [0.5, 0.5]
    >>> issolution(constraint, [.5,.5])
    True

    >>> from mystic.penalty import quadratic_inequality
    >>> @quadratic_inequality(lambda x: x[0] - x[1] + 10)
    ... def penalty(x):
    ...   return 0.0
    ... 
    >>> penalty([-10,.5])
    0.0
    >>> issolution(penalty, [-10,.5])
    True
    """
    if hasattr(constraints, 'error'):
        error = constraints.error(guess)
    else: # is a constraints solver
        try:
            constrained = guess.copy()
        except:
            from copy import deepcopy
            constrained = deepcopy(guess)
        error = 0.0
        constrained = constraints(constrained)
        for i in range(len(guess)):
            error += (constrained[i] - guess[i])**2  #XXX: better rnorm ?
        error = error**0.5

    return error <= tol


#XXX: nice if penalty.error could give error for each condition... or total


def solve(constraints, guess=None, nvars=None, solver=None, \
          lower_bounds=None, upper_bounds=None, termination=None, \
          tightrange=None, cliprange=None):
    """Use optimization to find a solution to a set of constraints.

Inputs:
    constraints -- a constraints solver function or a penalty function

Additional Inputs:
    guess -- list of parameter values proposed to solve the constraints.
    lower_bounds -- list of lower bounds on solution values.
    upper_bounds -- list of upper bounds on solution values.
    nvars -- number of parameter values.
    solver -- the mystic solver to use in the optimization.
    termination -- the mystic termination to use in the optimization.
    tightrange -- if True, impose bounds and constraints concurrently.
    cliprange -- if True, bounding constraints clip exterior values.

NOTE: The resulting constraints will likely be more expensive to evaluate
    and less accurate than writing the constraints solver from scratch.

NOTE: The ensemble solvers are available, using the default NestedSolver,
    where the keyword 'guess' can be used to set the number of solvers.

NOTE: The default solver is 'diffev', with npop=min(40, ndim*5). The default
    termination is ChangeOverGeneration(), and the default guess is randomly
    selected points between the upper and lower bounds.
    """
    npts = 8
    from numbers import Integral
    if isinstance(guess, Integral): npts, guess = guess, None

    ndim = 1 #XXX: better, increase in while loop catching IndexError ?
    if nvars is not None: ndim = nvars
    elif guess is not None: ndim = len(guess)
    elif lower_bounds is not None: ndim = len(lower_bounds)
    elif upper_bounds is not None: ndim = len(upper_bounds)

    def cost(x): return 1.

    #XXX: don't allow solver string as a short-cut?
    ensemble = False
    if solver is None or getattr(solver, '__name__', solver) == 'diffev':
        from mystic.solvers import DifferentialEvolutionSolver as TheSolver
        solver = TheSolver(ndim, min(40, ndim*5))
    elif getattr(solver, '__name__', solver) == 'diffev2':
        from mystic.solvers import DifferentialEvolutionSolver2 as TheSolver
        solver = TheSolver(ndim, min(40, ndim*5))
    #XXX: fmin_powell better as the default? (it's not random)
    elif getattr(solver, '__name__', solver) == 'fmin_powell':
        from mystic.solvers import PowellDirectionalSolver as TheSolver
        solver = TheSolver(ndim)
    elif getattr(solver, '__name__', solver) == 'fmin':
        from mystic.solvers import NelderMeadSimplexSolver as TheSolver
        solver = TheSolver(ndim)
    elif getattr(solver, '__name__', solver) == 'buckshot':
        from mystic.solvers import BuckshotSolver as TheSolver
        solver = TheSolver(ndim, max(8, npts)) #XXX: needs better default?
        ensemble = True
    elif getattr(solver, '__name__', solver) == 'lattice':
        from mystic.solvers import LatticeSolver as TheSolver
        solver = TheSolver(ndim, max(8, npts)) #XXX: needs better default?
        ensemble = True
    elif getattr(solver, '__name__', solver) == 'sparsity':
        from mystic.solvers import SparsitySolver as TheSolver
        solver = TheSolver(ndim, max(8, npts)) #XXX: needs better default?
        ensemble = True
    elif getattr(solver, '__name__', solver) == 'residual':
        from mystic.solvers import ResidualSolver as TheSolver
        solver = TheSolver(ndim, max(8, npts)) #XXX: needs better default?
        ensemble = True
    
    if termination is None:
        from mystic.termination import ChangeOverGeneration as COG
        termination = COG()
    if not ensemble:
        if guess is not None:
            solver.SetInitialPoints(guess) #XXX: nice if 'diffev' had methods
        else:
            solver.SetRandomInitialPoints(lower_bounds, upper_bounds)
    if lower_bounds or upper_bounds: #XXX: disable/hardwire tight, clip?
        kwds = dict(tight=tightrange, clip=cliprange)
        solver.SetStrictRanges(lower_bounds, upper_bounds, **kwds)
    if hasattr(constraints, 'iter') and hasattr(constraints, 'error'):
        solver.SetPenalty(constraints) #i.e. is a penalty function
    else: # is a constraints solver
        solver.SetConstraints(constraints)
    from numpy import seterr
    settings = seterr(all='ignore')
    solver.Solve(cost, termination)
    seterr(**settings)
    soln = solver.bestSolution

    from numpy import ndarray, array
    if isinstance(soln, ndarray) and not isinstance(guess, ndarray):
        soln = soln.tolist()
    elif isinstance(guess, ndarray) and not isinstance(soln, ndarray):
        soln = array(soln)  #XXX: or always return a list ?

    return soln #XXX: check with 'issolution' ?


def as_constraint(penalty, *args, **kwds):
    """Convert a penalty function to a constraints solver.

Inputs:
    penalty -- a penalty function

Additional Inputs:
    lower_bounds -- list of lower bounds on solution values.
    upper_bounds -- list of upper bounds on solution values.
    nvars -- number of parameter values.
    solver -- the mystic solver to use in the optimization.
    termination -- the mystic termination to use in the optimization.
    tightrange -- if True, impose bounds and constraints concurrently.
    cliprange -- if True, bounding constraints clip exterior values.

NOTE: The default solver is 'diffev', with npop=min(40, ndim*5). The default
    termination is ChangeOverGeneration(), and the default guess is randomly
    selected points between the upper and lower bounds.
    """
    def constraint(x): #XXX: better to enable args kwds for penalty ?
        return solve(penalty, x, *args, **kwds)
    return constraint


def as_penalty(constraint, ptype=None, *args, **kwds):
    """Convert a constraints solver to a penalty function.

Inputs:
    constraint -- a constraints solver
    ptype -- penalty function type [default: quadratic_equality]

Additional Inputs:
    args -- arguments for the constraints solver [default: ()]
    kwds -- keyword arguments for the constraints solver [default: {}]
    k -- penalty multiplier
    h -- iterative multiplier
    """
    def rnorm(x, *argz, **kwdz):
        error = 0.0
        constrained = constraint(x, *argz, **kwdz)
        for i in range(len(x)):
            error += (constrained[i] - x[i])**2  #XXX: better rnorm ?
        error = error**0.5
        return error

    if ptype is None:
        from mystic.penalty import quadratic_equality
        ptype = quadratic_equality

    @ptype(rnorm, *args, **kwds) #XXX: yes to k,h... but otherwise?
    def penalty(x):
        return 0.0

    penalty.func = rnorm
    penalty.ptype = ptype.__name__ 
    return penalty


# constraints 'language'
#XXX: should have the same interface as 'inner' and 'outer'?
def and_(*constraints, **settings): #XXX: not a decorator, should be?
    """combine several constraints into a single constraint

Inputs:
    constraints -- constraint functions

Additional Inputs:
    maxiter -- maximum number of iterations to attempt to solve [default: 100]
    onexit -- function x' = f(x) to call on success [default: None]
    onfail -- function x' = f(x) to call on failure [default: None]

NOTE: 
    If a repeating cycle is detected, some of the inputs may be randomized.
    """
    import itertools as it
    import random as rnd
    n = len(constraints)
    if 'maxiter' in settings:
        maxiter = settings['maxiter'] * n; del settings['maxiter']
    else: maxiter = 100 * n
    if 'onexit' in settings:
        onexit = settings['onexit']; del settings['onexit']
    else: onexit = None
    if 'onfail' in settings:
        onfail = settings['onfail']; del settings['onfail']
    else: onfail = None
    def _constraint(x): #XXX: inefficient, rewrite without append
        x = [x.tolist() if hasattr(x, 'tolist') else x[:]]
        # apply all constraints once
        e = None
        for c in constraints:
            try:
                ci = c(x[-1][:])
            except ZeroDivisionError as exc:
                e = exc
                ci = x[-1][:] #XXX: do something else?
            except (TypeError, ValueError) as exc:
                if (exc.args[0].find('not supported') and \
                    exc.args[0].rfind("'complex'")):
                    e = exc
                    ci = x[-1][:] #XXX: do something else?
                else: raise exc
            x.append(ci.tolist() if hasattr(ci, 'tolist') else ci)
        if all(xi == x[-1] for xi in x[1:]) and e is None:
            return x[-1] if onexit is None else onexit(x[-1][:])
        # cycle constraints until there's no change
        _constraints = it.cycle(constraints) 
        for j in range(n,maxiter):
            e = None
            try:
                ci = next(_constraints)(x[-1][:])
            except ZeroDivisionError as exc:
                e = exc
                ci = x[-1][:] #XXX: do something else?
            except (TypeError, ValueError) as exc:
                if (exc.args[0].find('not supported') and \
                    exc.args[0].rfind("'complex'")):
                    e = exc
                    ci = x[-1][:] #XXX: do something else?
                else: raise exc
            x.append(ci.tolist() if hasattr(ci, 'tolist') else ci)
            if all(xi == x[-1] for xi in x[-n:]) and e is None:
                return x[-1] if onexit is None else onexit(x[-1][:])
            # may be trapped in a cycle... randomize
            if x[-1] == x[-(n+1)]:
                x[-1] = [(i+rnd.randint(-1,1))*rnd.random() for i in x[-1]]
            if not j%(2*n):
                del x[:n]
        # give up #XXX: or fail with Error?
        return x[-1] if onfail is None else onfail(x[-1][:])
    cf = lambda x: _constraint(x)
    cfdoc = "\n-- AND --\n".join(c.__doc__ for c in constraints if c.__doc__)
    cfdoc = '{ '+ cfdoc +' }' if cfdoc else cfdoc #XXX: can be {c} w/no AND
    cf.__doc__ = cfdoc.rstrip('\n')
    cf.__name__ = 'constraint'
    return cf


#XXX: should have the same interface as 'inner' and 'outer'?
def or_(*constraints, **settings): #XXX: not a decorator, should be?
    """create a constraint that is satisfied if any constraints are satisfied

Inputs:
    constraints -- constraint functions

Additional Inputs:
    maxiter -- maximum number of iterations to attempt to solve [default: 100]
    onexit -- function x' = f(x) to call on success [default: None]
    onfail -- function x' = f(x) to call on failure [default: None]

NOTE: 
    If a repeating cycle is detected, some of the inputs may be randomized.
    """
    import itertools as it
    import random as rnd
    n = len(constraints)
    if 'maxiter' in settings:
        maxiter = settings['maxiter'] * n; del settings['maxiter']
    else: maxiter = 100 * n
    if 'onexit' in settings:
        onexit = settings['onexit']; del settings['onexit']
    else: onexit = None
    if 'onfail' in settings:
        onfail = settings['onfail']; del settings['onfail']
    else: onfail = None
    def _constraint(x): #XXX: inefficient, rewrite without append
        x = [x.tolist() if hasattr(x, 'tolist') else x[:]]
        # check if initial input is valid
        e = None
        for c in constraints:
            try:
                ci = c(x[0][:])
            except ZeroDivisionError as exc:
                e = exc
                ci = x[0][:] #XXX: do something else?
            except (TypeError, ValueError) as exc:
                if (exc.args[0].find('not supported') and \
                    exc.args[0].rfind("'complex'")):
                    e = exc
                    ci = x[-1][:] #XXX: do something else?
                else: raise exc
            x.append(ci.tolist() if hasattr(ci, 'tolist') else ci)
            if x[-1] == x[0] and e is None:
                return x[-1] if onexit is None else onexit(x[-1][:])
        # cycle constraints until there's no change
        _constraints = it.cycle(constraints) 
        for j in range(n,maxiter):
            e = None
            try:
                ci = next(_constraints)(x[-n][:])
            except ZeroDivisionError as exc:
                e = exc
                ci = x[-n][:] #XXX: do something else?
            except (TypeError, ValueError) as exc:
                if (exc.args[0].find('not supported') and \
                    exc.args[0].rfind("'complex'")):
                    e = exc
                    ci = x[-1][:] #XXX: do something else?
                else: raise exc
            x.append(ci.tolist() if hasattr(ci, 'tolist') else ci)
            if x[-1] == x[-(n+1)] and e is None:
                return x[-1] if onexit is None else onexit(x[-1][:])
            else: # may be trapped in a rut... randomize
                x[-1] = x[-rnd.randint(1,n)]
            if not j%(2*n):
                del x[:n]
        # give up #XXX: or fail with Error?
        return x[-1] if onfail is None else onfail(x[-1][:])
    cf = lambda x: _constraint(x)
    cfdoc = "\n-- OR --\n".join(c.__doc__ for c in constraints if c.__doc__)
    cfdoc = '[ '+ cfdoc +' ]' if cfdoc else cfdoc #XXX: can be [c] w/no OR
    cf.__doc__ = cfdoc.rstrip('\n')
    cf.__name__ = 'constraint'
    return cf


def not_(constraint, **settings): #XXX: not a decorator, should be?
    """invert the region where the given constraints are valid, then solve

Inputs:
    constraint -- constraint function

Additional Inputs:
    maxiter -- maximum number of iterations to attempt to solve [default: 100]
    onexit -- function x' = f(x) to call on success [default: None]
    onfail -- function x' = f(x) to call on failure [default: None]

NOTE: 
    If a repeating cycle is detected, some of the inputs may be randomized.
    """
    import random as rnd
    if 'maxiter' in settings:
        maxiter = settings['maxiter']; del settings['maxiter']
    else: maxiter = 100
    if 'onexit' in settings:
        onexit = settings['onexit']; del settings['onexit']
    else: onexit = None
    if 'onfail' in settings:
        onfail = settings['onfail']; del settings['onfail']
    else: onfail = None
    def _constraint(x):
        # check if initial input is valid, else randomize and try again
        for j in range(0,maxiter):
            try:
                if constraint(x[:]) != x:
                    return x[:] if onexit is None else onexit(x[:])
            except ZeroDivisionError as e:
                pass #XXX: do something else?
            except (TypeError, ValueError) as exc:
                if (exc.args[0].find('not supported') and \
                    exc.args[0].rfind("'complex'")):
                    pass #XXX: do something else?
                else: raise exc
            x = [(i+rnd.randint(-1,1))*rnd.random() for i in x]
        # give up #XXX: or fail with Error?
        return x[:] if onfail is None else onfail(x[:])
    cf = lambda x: _constraint(x)
    cfdoc = 'NOT( '+ constraint.__doc__ +' )' if constraint.__doc__ else ""
    cf.__doc__ = cfdoc.rstrip('\n')
    cf.__name__ = 'constraint'
    return cf


from numpy import asarray, choose, zeros, ones, ndarray
from numpy import shape, broadcast, empty, atleast_1d
#from random import sample, choice
def discrete(samples, index=None):
    """impose a discrete set of input values for the selected function

The function's input will be mapped to the given discrete set

Examples:
    >>> @discrete([1.0, 2.0])
    ... def identity(x):
    ...     return x
    ...
    >>> identity([0.123, 1.789, 4.000])
    [1.0, 2.0, 2.0]

    >>> @discrete([1,3,5,7], index=(0,3))
    ... def squared(x):
    ...     return [i**2 for i in x]
    ...
    >>> squared([0,2,4,6,8,10])
    [1, 4, 16, 25, 64, 100]
    """
    samples = [asarray(samples)]
    samples[0].sort()
    from numbers import Integral
    if isinstance(index, Integral): index = (index,)
    index = [index]

    #XXX: refactor to use points_factory(samples)
    def _points(alist):
        alist = asarray(alist)
        alist.sort()
        samples[0] = alist

    def _index(alist=None):
        index[0] = alist

    #XXX: refactor to use argnear_factory(samples)
    def _argnear(xi):
        arghi = sum(xi > samples[0])
        arglo = max(0, arghi - 1) # minimum = x[0]
        if arghi == len(samples[0]): 
            arghi = arglo         # maximum = x[-1]
        return arglo, arghi

    def _near(xi, lo, hi):
        if hi - xi < xi - lo: 
            return hi
        return lo

#   argnear = vectorize(_argnear)  #XXX: doesn't pickle well in all cases
#   near = vectorize(_near)        #XXX: doesn't pickle well in all cases
    def argnear(x):
        #RESULT: (array([0,0,1]),array([0,1,2])) *or* (array(0),array(1))
        flatten = False
        if not len(shape(x)):
            flatten = True
            x = [x]
        result = tuple(i for i in asarray(list(map(_argnear, x))).T)
        if flatten:
            result = tuple(asarray(i[0]) for i in result)
        return result
    def near(x, l, h):
        #RESULT: array(3) & array([3, 4])
        a = broadcast(x,l,h)
        has_iterable = a.shape
        x,l,h = tuple(atleast_1d(i) for i in (x,l,h))
        b = broadcast(x,l,h)
        _x,_l,_h = (empty(b.shape), empty(b.shape), empty(b.shape))
        _x.flat = [i for i in x]
        _l.flat = [i for i in l]
        _h.flat = [i for i in h]
        result = asarray(list(map(_near, x, l, h)))
        if not has_iterable:
            result = asarray(result[0])
        return result

    def dec(f):
        def func(x, *args, **kwds):
            if isinstance(x, ndarray): xtype = asarray
            else: xtype = type(x)
            arglo, arghi = argnear(x)
            xp = near(x, samples[0][arglo], samples[0][arghi])
            # create a choice array from given indices
            #FIXME: better ways to do the following
            if index[0] is None: 
                mask = ones(xp.size, dtype=bool)
            else:
                mask = zeros(xp.size, dtype=bool)
                idx = filter(lambda i:-xp.size<=i<=xp.size-1, index[0])
                try: mask[sorted(idx, key=abs)] = True
                except IndexError: pass
            xp = choose(mask, (x,xp)).reshape(xp.shape)
            return f(xtype(xp), *args, **kwds)
        func.samples = _points
        func.index = _index
        return func
    return dec


from numpy import round, abs
def integers(ints=True, index=None):
    """impose the set of integers (by rounding) for the given function

The function's input will be mapped to the ints, where:
  - if ints is True, return results as ints; otherwise, use floats
  - if index tuple provided, only round at the given indices

Examples:
  >>> @integers()
  ... def identity(x):
  ...     return x
  ...
  >>> identity([0.123, 1.789, 4.000])
  [0, 2, 4]

  >>> @integers(ints=float, index=(0,3,4))
  ... def squared(x):
  ...     return [i**2 for i in x]
  ...
  >>> squared([0.12, 0.12, 4.01, 4.01, 8, 8])
  [0.0, 0.0144, 16.080099999999998, 16.0, 64.0, 64.0]
    """
    #HACK: allow ints=False or ints=int
    _ints = [(int if ints else float) if (repr(ints) in ('True', 'False')) else ints]
    from numbers import Integral
    if isinstance(index, Integral): index = (index,)
    index = [index]

    def _index(alist=None):
        index[0] = alist

    def _type(ints=None):
        _ints[0] = (int if ints else float) if (repr(ints) in ('True', 'False')) else ints

    def dec(f):
        def func(x,*args,**kwds):
            if isinstance(x, ndarray): xtype = asarray
            else: xtype = type(x)
            ##### ALT #####
           #pos = range(len(x))
           #pos = [pos[i] for i in index[0]] if index[0] else pos
           #xp = [float(round(xi) if i in pos else xi) for i,xi in enumerate(x)]
            ###############
            xp = round(x)
            if index[0] is None:
                mask = ones(xp.size, dtype=bool)
            else:
                mask = zeros(xp.size, dtype=bool)
                idx = filter(lambda i:-xp.size<=i<=xp.size-1, index[0])
                try: mask[sorted(idx, key=abs)] = True
                except IndexError: pass
            xp = choose(mask, (x,xp)).astype(_ints[0]).reshape(xp.shape)
            ###############
            return f(xtype(xp), *args, **kwds)
        func.index = _index
        func.type = _type
        return func
    return dec


def rounded(digits=None, index=None):
    """rounded inputs for the given function

The function's input will be mapped to the given precision, where:
  - digits is an int that sets the rounding precision (can be negative)
  - if index tuple provided, only round at the given indices

Examples:
  >>> @rounded()
  ... def identity(x):
  ...     return x
  ...
  >>> identity([0.123, 1.789, 4.000])
  [0.0, 2.0, 4.0]

  >>> @rounded(digits=1, index=(0,3,4))
  ... def squared(x):
  ...     return [i**2 for i in x]
  ...
  >>> squared([123.45, 123.45, 4.01, 4.01, 0.012, 0.012])
  [15227.560000000001, 15239.9025, 16.080099999999998, 16.0, 0.0, 0.000144]

  >>> @rounded(digits=-1, index=(0,3,4))
  ... def square(x):
  ...     return [i**2 for i in x]
  ...
  >>> square([123.45, 123.45, 4.01, 4.01, 0.012, 0.012])
  [14400.0, 15239.9025, 16.080099999999998, 0.0, 0.0, 0.000144]
    """
    digits = [digits or 0]
    from numbers import Integral
    if isinstance(index, Integral): index = (index,)
    index = [index]

    def _index(alist=None):
        index[0] = alist

    def _digits(aint=None):
        digits[0] = aint or 0

    def dec(f):
        def func(x,*args,**kwds):
            if isinstance(x, ndarray): xtype = asarray
            else: xtype = type(x)
            xp = round(x, digits[0])
            if index[0] is None:
                mask = ones(xp.size, dtype=bool)
            else:
                mask = zeros(xp.size, dtype=bool)
                idx = filter(lambda i:-xp.size<=i<=xp.size-1, index[0])
                try: mask[sorted(idx, key=abs)] = True
                except IndexError: pass
            xp = choose(mask, (x,xp)).astype(float).reshape(xp.shape)
            return f(xtype(xp), *args, **kwds)
        func.index = _index
        func.digits = _digits
        return func
    return dec


def precision(digits=None, index=None):
    """rounded outputs for the given function

The function's output will be mapped to the given precision, where:
  - digits is an int that sets the rounding precision (can be negative)
  - if index tuple provided, only round at the given indices

Examples:
  >>> @precision()
  ... def identity(x):
  ...     return x
  ...
  >>> identity([0.123, 1.789, 4.000])
  [0.0, 2.0, 4.0]

  >>> @precision(digits=1, index=(0,3,4))
  ... def squared(x):
  ...     return [i**2 for i in x]
  ...
  >>> squared([123.45, 123.45, 4.01, 4.01, 0.012, 0.012])
  [15239.9, 15239.9025, 16.080099999999998, 16.1, 0.0, 0.000144]

  >>> @precision(digits=-1, index=(0,3,4))
  ... def square(x):
  ...     return [i**2 for i in x]
  ...
  >>> square([123.45, 123.45, 4.01, 4.01, 0.012, 0.012])
  [15240.0, 15239.9025, 16.080099999999998, 20.0, 0.0, 0.000144]
    """
    digits = [digits or 0]
    from numbers import Integral
    if isinstance(index, Integral): index = (index,)
    index = [index]

    def _index(alist=None):
        index[0] = alist

    def _digits(aint=None):
        digits[0] = aint or 0

    def dec(f):
        def func(x,*args,**kwds):
            fx = f(x, *args, **kwds)
            if isinstance(fx, ndarray): xtype = asarray
            else: xtype = type(fx)
            y = round(fx, digits[0])
            if index[0] is None:
                mask = ones(y.size, dtype=bool)
            else:
                mask = zeros(y.size, dtype=bool)
                idx = filter(lambda i:-y.size<=i<=y.size-1, index[0])
                try: mask[sorted(idx, key=abs)] = True
                except IndexError: pass
            y = choose(mask, (fx,y)).reshape(y.shape)#.astype(float)
            return xtype(y)
        func.index = _index
        func.digits = _digits
        return func
    return dec


from random import shuffle, Random
# subclass to get '_randbelow_without_getrandbits' behavior for randrange
class _Random(Random):
    def _randbelow(self, n):
        from random import random as _random
        return n * _random()

    def randrange(self, start, stop=None, step=1, _int=int):
        """Choose a random item from range(start, stop[, step]).
        """ #NOTE: from python3.8/random.py^
        # doing adequate error checking.

        istart = _int(start)
        if istart != start:
            raise ValueError("non-integer arg 1 for randrange()")
        if stop is None:
            if istart > 0:
                return self._randbelow(istart)
            raise ValueError("empty range for randrange()")

        # stop argument supplied.
        istop = _int(stop)
        if istop != stop:
            raise ValueError("non-integer stop for randrange()")
        width = istop - istart
        if step == 1 and width > 0:
            return istart + self._randbelow(width)
        if step == 1:
            raise ValueError("empty range for randrange() (%d, %d, %d)" % (istart, istop, width))

        # Non-unit step argument supplied.
        istep = _int(step)
        if istep != step:
            raise ValueError("non-integer step for randrange()")
        if istep > 0:
            n = (width + istep - 1) // istep
        elif istep < 0:
            n = (width + istep + 1) // istep
        else:
            raise ValueError("zero step for randrange()")

        if n <= 0:
            raise ValueError("empty range for randrange()")

        return istart + istep*self._randbelow(n)

randrange = _Random().randrange
del _Random, Random

def unique(seq, full=None):
    """replace the duplicate values with unique values in 'full'

    If full is a type (int or float), then unique values of the given type
    are selected from range(min(seq),max(seq)). If full is a dict of
    {'min':min, 'max':max}, then unique floats are selected from
    range(min(seq),max(seq)). If full is a sequence (list or set), then
    unique values are selected from the given sequence. 

    Examples:
      >>> unique([1,2,3,1,2,4], range(11))
      [1, 2, 3, 9, 8, 4]

      >>> unique([1,2,3,1,2,9], range(11))
      [1, 2, 3, 8, 5, 9]

      >>> try:
      ...     unique([1,2,3,1,2,13], range(11))
      ...     raise Exception
      ... except ValueError:
      ...     pass
      ...

      >>> unique([1,2,3,1,2,4], {'min':0, 'max':11})
      [1, 2, 3, 4.175187820357143, 2.5407265707465716, 4]

      >>> unique([1,2,3,1,2,4], {'min':0, 'max':11, 'type':int})
      [1, 2, 3, 6, 8, 4]

      >>> unique([1,2,3,1,2,4], float)
      [1, 2, 3, 1.012375036824941, 3.9821250727509905, 4]

      >>> unique([1,2,3,1,2,10], int)
      [1, 2, 3, 9, 6, 10]

      >>> try:
      ...     unique([1,2,3,1,2,4], int)
      ...     raise Exception
      ... except ValueError:
      ...     pass
      ...
    """ #FIXME: does not handle full is numpy.dtype (i.e. int64, float64)
    # check type if full not specified
    if full is None:
        from numbers import Integral
        if all([isinstance(x, Integral) for x in seq]): full = int
        else: full = float
        ok = True
    else: ok = False
    # replace all duplicates with 'None'
    unique = set()
    seq = [x if x not in unique and not unique.add(x) else None for x in seq]
    lseq = len(seq)
    # check all unique show up in 'full'
    if full in (int,float): # specified type, not range
        from numbers import Integral
        ok = ok or full==float or all([isinstance(x, Integral) for x in unique])
        msg = "not all items are of type='%s'" % full.__name__
        _min = min(unique)
        _max = max(unique)
    elif isinstance(full, dict): # specified min/max for floats
        if 'type' in full: #NOTE: undocumented keys: min,max,type
            _type = full['type']; del full['type']
        else: _type = float
        minu = min(unique)
        maxu = max(unique)
        _min = full['min'] if 'min' in full else minu
        _max = full['max'] if 'max' in full else maxu
        ok = minu >= _min and maxu < _max
        if not ok:
            oops = list(unique - set(range(_min,_max)))
            msg = "x=%s not in %s <= x < %s" % (oops[-1],_min,_max)
        full = _type
    else: # full is a list of all possible values
        ok = unique.issubset(full)
        if not ok:
            oops = list(unique - set(full))
            msg = "%s not in given set" % oops[-1]
        _min = _max = None
    if not ok: raise ValueError(msg)
    # check if a unique sequence is possible to build
    if full is float:
        if _min == _max and lseq > 1:
            msg = "no unique len=%s sequence with %s <= x <= %s" % (lseq,_min,_max)
            raise ValueError(msg)
        # replace the 'None' values in seq with 'new' values
        #XXX: HIGHLY UNLIKELY two numbers will be the same, but possible
        return [randrange(_min,_max,_int=float) if x is None else x for x in seq]
    # generate all possible values not found in 'unique'
    if full is int:
        if max(lseq - (_max+1 - _min), 0):
            msg = "no unique len=%s sequence with %s <= x <= %s" % (lseq,_min,_max)
            raise ValueError(msg)
        new = list(set(range(_min,_max+1)) - unique)
    else:
        if lseq > len(full):
            msg = "no unique len=%s sequence in given set" % lseq
            raise ValueError(msg)
        new = list(set(full) - unique)
    # ensure randomly ordered
    shuffle(new)
    # replace the 'None' values in seq with 'new' values
    return [new.pop() if x is None else x for x in seq]


#XXX: enable impose_unique on selected members of x? (see constraints.integers)

def impose_unique(seq=None):
    """ensure all values are unique and found in the given set

    Examples:
      >>> @impose_unique(range(11))
      ... def doit(x):
      ...     return x
      ... 
      >>> doit([1,2,3,1,2,4])
      [1, 2, 3, 9, 8, 4]
      >>> doit([1,2,3,1,2,10])
      [1, 2, 3, 8, 5, 10]
      >>> try:
      ...     doit([1,2,3,1,2,13])
      ...     raise Exception
      ... except ValueError:
      ...     pass
      ...
"""
    def dec(f):
        def func(x,*args,**kwds):
            return f(unique(x, seq),*args,**kwds)
        return func
    return dec

from numpy import array, intersect1d, inf, isnan, where, choose, clip as _clip
from numpy.random import uniform, choice
def bounded(seq, bounds, index=None, clip=True, nearest=True):
    """bound a sequence by bounds = [min,max]

    Examples:
      >>> sequence = [0.123, 1.244, -4.755, 10.731, 6.207]
      >>> 
      >>> bounded(sequence, (0,5))
      array([0.123, 1.244, 0.   , 5.   , 5.   ])
      >>> 
      >>> bounded(sequence, (0,5), index=(0,2,4))
      array([ 0.123,  1.244,  0.   , 10.731,  5.   ])
      >>> 
      >>> bounded(sequence, (0,5), clip=False)
      array([0.123     , 1.244     , 3.46621839, 1.44469038, 4.88937466])
      >>> 
      >>> bounds = [(0,5),(7,10)]
      >>> bounded(sequence, bounds)
      array([ 0.123,  1.244,  0.   , 10.   ,  7.   ])
      >>>
      >>> bounded(sequence, bounds, nearest=False)
      array([ 0.123,  1.244,  7.   , 10.   ,  5.   ])
      >>>
      >>> bounded(sequence, bounds, nearest=False, clip=False) 
      array([0.123     , 1.244     , 0.37617154, 8.79013111, 7.40864242])
      >>>
      >>> bounded(sequence, bounds, clip=False)
      array([0.123     , 1.244     , 2.38186577, 7.41374049, 9.14662911])
"""
    seq = array(seq) #XXX: asarray?
    if bounds is None or not bounds: return seq
    from numbers import Integral
    if isinstance(index, Integral): index = (index,)
    if not hasattr(bounds[0], '__len__'): bounds = (bounds,)
    bounds = asarray(bounds, dtype='float64').T # [(min,min,...),(max,max,...)]
    # convert None to -inf or inf
    bounds[0][isnan(bounds[0])] = -inf
    bounds[1][isnan(bounds[1])] = inf
    # find indicies of the elements that are out of bounds
    at = where(sum((lo <= seq)&(seq <= hi) for (lo,hi) in bounds.T).astype(bool) == False)[-1]
    # find the intersection of out-of-bound and selected indicies
    at = at if index is None else intersect1d(at, index)
    if not len(at): return seq
    if clip:
        if nearest: # clip at closest bounds
            seq_at = seq[at]
            seq[at] = _clip(seq_at, *(b[abs(seq_at.reshape(-1,1)-b).argmin(axis=1)] for b in bounds))
        else: # clip in randomly selected interval
            picks = choice(len(bounds.T), size=at.shape)
            seq[at] = _clip(seq[at], bounds[0][picks], bounds[1][picks])
        return seq
    # limit to +/- 1e300 #XXX: better defaults?
    bounds[0][bounds[0] < -1e300] = -1e300
    bounds[1][bounds[1] > 1e300] = 1e300
    if nearest:
        seq_at = seq[at]
        seq[at] = choose(array([abs(seq_at.reshape(-1,1) - b).min(axis=1) for b in bounds.T]).argmin(axis=0), [uniform(0,1, size=at.shape) * (hi - lo) + lo for (lo,hi) in bounds.T])
    else: # randomly choose a value in one of the intervals
        seq[at] = choose(choice(len(bounds.T), size=at.shape), [uniform(0,1, size=at.shape) * (hi - lo) + lo for (lo,hi) in bounds.T])
    return seq

def impose_bounds(bounds, index=None, clip=True, nearest=True):
    """generate a function where bounds=[min,max] on a sequence are imposed

    Examples:
      >>> sequence = [0.123, 1.244, -4.755, 10.731, 6.207]
      >>> 
      >>> @impose_bounds((0,5))       
      ... def simple(x):
      ...   return x
      ... 
      >>> simple(sequence)
      [0.123, 1.244, 0.0, 5.0, 5.0]
      >>> 
      >>> @impose_bounds((0,5), index=(0,2,4))
      ... def double(x):
      ...   return [i*2 for i in x]
      ... 
      >>> double(sequence)
      [0.246, 2.488, 0.0, 21.462, 10.0]
      >>> 
      >>> @impose_bounds((0,5), index=(0,2,4), clip=False)
      ... def square(x):
      ...   return [i*i for i in x]
      ... 
      >>> square(sequence)
      [0.015129, 1.547536, 14.675791119810688, 115.154361, 1.399551896073788]
      >>> 
      >>> @impose_bounds([(0,5),(7,10)])
      ... def simple(x):
      ...   return x
      ... 
      >>> simple(sequence)
      [0.123, 1.244, 0.0, 10.0, 7.0]
      >>> 
      >>> @impose_bounds([(0,5),(7,10)], nearest=False)
      ... def simple(x):
      ...   return x
      ... 
      >>> simple(sequence)
      [0.123, 1.244, 0.0, 5.0, 5.0]
      >>> simple(sequence)
      [0.123, 1.244, 7.0, 10.0, 5.0]
      >>> 
      >>> @impose_bounds({0:(0,5), 2:(0,5), 4:[(0,5),(7,10)]})
      ... def simple(x):
      ...   return x
      ... 
      >>> simple(sequence)
      [0.123, 1.244, 0.0, 10.731, 7.0]
      >>>
      >>> @impose_bounds({0:(0,5), 2:(0,5), 4:[(0,5),(7,10)]}, index=(0,2)) 
      ... def simple(x):
      ...   return x
      ... 
      >>> simple(sequence)
      [0.123, 1.244, 0.0, 10.731, 6.207]
    """
    ### bounds is a dict, index filters the dict
    ### keys of bounds are the index, if index=None apply to all
    ### *but* if bounds is given as list (of tuples), apply to seleted index
    from numbers import Integral
    if isinstance(index, Integral): index = (index,)
    # bounds are list, index is tuple => {i:bounds} for each i in index
    if type(bounds) is not dict:
        if index is not None:
            bounds = dict((i,bounds) for i in index)
            index = None
        # bounds are list, index is None => {None:bounds} apply to all index
        else: bounds = {None:bounds}
    # bounds are dict, index is None => do nothing -- is the preferred input
    # bounds are dict, index is tuple => filter bounds, then as above
    else:
        if None in bounds and len(bounds) > 1: # has either entries or None
            msg = "bounds got multiple entries for '%d'" % set(bounds).difference((None,)).pop()
            raise TypeError(msg)
        if index is not None:
            if None in index and len(index) > 1: # has either entries or None
                msg = "index got multiple entries for '%d'" % set(index).difference((None,)).pop()
                raise TypeError(msg)
            if None in bounds:
                bounds = dict((i,bounds[None]) for i in index)
            else:
                index = set(index).intersection(bounds)
                bounds = dict((i,bounds[i]) for i in index)
            index = None

    clip = [clip]
    nearest = [nearest]

    def _clip(clipped=True):
        clip[0] = clipped

    def _near(nearer=True):
        nearest[0] = nearer

    def dec(f):
        def func(x,*args,**kwds):
            if isinstance(x, ndarray): xtype = asarray
            else: xtype = type(x)
            xp = x
            for i in bounds:
                xp = bounded(xp, bounds[i], i, clip[0], nearest[0])
            return f(xtype(xp), *args, **kwds)
        func.clip = _clip
        func.nearest = _near
        func.__bounds__ = bounds
        func.__wrapped__ = f
        func.__doc__ = f.__doc__
        return func
    return dec


from numpy import maximum, minimum
from operator import itemgetter
def monotonic(ascending=True, outer=False, index=None):
    """impose monotonicity on the given function

The function will become monotonic, where:
  - ascending is bool, increasing with index if True and decreasing if False
  - outer is bool, applied to the outputs if True and to the inputs if False
  - if index is given, only apply at the given sequence defined by index

Examples:
  >>> @monotonic(outer=True)
  ... def negative(x):
  ...     return [-i for i in x]
  ...
  >>> negative([5.34, -.121, -4.11, 9.01, 11.3, -16.4])
  [-5.34, 0.121, 4.11, 4.11, 4.11, 16.4]
  >>> 
  >>> @monotonic()
  ... def negative(x):
  ...     return [-i for i in x]
  ...
  >>> negative([5.34, -.121, -4.11, 9.01, 11.3, -16.4])
  [-5.34, -5.34, -5.34, -9.01, -11.3, -11.3]
  >>> 
  >>> @monotonic(outer=True)
  ... def squared(x):
  ...     return [i*i for i in x]
  ...
  >>> squared([5.34, -.121, -4.11, 9.01, 11.3, -16.4])
  [28.5156, 28.5156, 28.5156, 81.1801, 127.69000000000001, 268.96]
  >>> 
  >>> @monotonic()
  ... def squared(x):
  ...     return [i*i for i in x]
  ...
  >>> squared([5.34, -.121, -4.11, 9.01, 11.3, -16.4])
  [28.5156, 28.5156, 28.5156, 81.1801, 127.69000000000001, 127.69000000000001]
  >>>
  >>> @monotonic(index=(0,1))
  ... def squared(x):
  ...   return [i*i for i in x]
  ...
  >>> squared([5.34, -.121, -4.11, 9.01, 11.3, -16.4])
  [28.5156, 28.5156, 16.892100000000003, 81.1801, 127.69000000000001, 268.96]
    """
    index = [index]

    def _index(alist=None):
        index[0] = alist

    def _mono(x, ascending=True):
        if not hasattr(x, '__len__'): return x
        if isinstance(x, ndarray): xtype = asarray
        else: xtype = type(x)
        fun = maximum.accumulate if ascending else minimum.accumulate
        return xtype(fun(x))

    def _imono(x, ascending=True):
        idx = index[0]
        if idx is None: return _mono(x, ascending=ascending)
        if not hasattr(idx, '__len__') or len(idx)==1: return x
        if not hasattr(x, '__len__') or len(x)==1: return x
        idx = sorted(itemgetter(*idx)(range(len(x))))
        for i,j in zip(idx,_mono(itemgetter(*idx)(x), ascending=ascending)):
            x[i] = j
        return x

    def dec(f):
        def func(x, *args, **kwds):
            if outer:
                return _imono(f(x, *args, **kwds), ascending=ascending)
            if index[0] is not None:
                try:
                    x = x.copy() #NOTE: assume shouldn't modify x
                except AttributeError:
                    import copy
                    x = copy.copy(x) #XXX: inefficient
            return f(_imono(x, ascending=ascending), *args, **kwds)
        func.index = _index
        func.monotonic = _imono
        func.__wrapped__ = f
        func.__doc__ = f.__doc__
        return func
    return dec


def sorting(ascending=True, outer=False, index=None):
    """impose sorting on the given function

The function will use sorting, where:
  - ascending is bool, increasing with index if True and decreasing if False
  - outer is bool, applied to the outputs if True and to the inputs if False
  - if index is given, only apply at the given sequence defined by index

Examples:
  >>> @sorting(outer=True)
  ... def squared(x):
  ...     return [i*i for i in x]
  ...
  >>> squared([5.34, -.121, -4.11, 9.01, 11.3, -16.4])
  [0.014641, 16.892100000000003, 28.5156, 81.1801, 127.69000000000001, 268.96]
  >>> 
  >>> @sorting()
  ... def squared(x):
  ...     return [i*i for i in x]
  ...
  >>> squared([5.34, -.121, -4.11, 9.01, 11.3, -16.4])
  [268.96, 16.892100000000003, 0.014641, 28.5156, 81.1801, 127.69000000000001]
  >>>
  >>> @sorting(index=(0,1,2))
  ... def squared(x):
  ...   return [i*i for i in x]
  ...
  >>> squared([5.34, -.121, -4.11, 9.01, 11.3, -16.4])
  [16.892100000000003, 0.014641, 28.5156, 81.1801, 127.69000000000001, 268.96]
    """
    from numbers import Integral
    if isinstance(index, Integral): index = (index,)
    index = [index]

    def _index(alist=None):
        index[0] = alist

    def _sort(x, ascending=True):
        if not hasattr(x, '__len__'): return x
        if isinstance(x, ndarray): xtype = asarray
        else: xtype = type(x)
        return xtype(sorted(x, reverse=(not ascending)))

    def _isort(x, ascending=True):
        idx = index[0]
        if idx is None: return _sort(x, ascending=ascending)
        if not hasattr(idx, '__len__') or len(idx)==1: return x
        if not hasattr(x, '__len__') or len(x)==1: return x
        idx = sorted(itemgetter(*idx)(range(len(x))))
        for i,j in zip(idx,_sort(itemgetter(*idx)(x), ascending=ascending)):
            x[i] = j
        return x

    def dec(f):
        def func(x, *args, **kwds):
            if outer:
                return _isort(f(x, *args, **kwds), ascending=ascending)
            if index[0] is not None:
                try:
                    x = x.copy() #NOTE: assume shouldn't modify x
                except AttributeError:
                    import copy
                    x = copy.copy(x) #XXX: inefficient
            return f(_isort(x, ascending=ascending), *args, **kwds)
        func.index = _index
        func.sorting = _isort
        func.__wrapped__ = f
        func.__doc__ = f.__doc__
        return func
    return dec


def periodic(period=None, index=None):
    """impose periodicity on the selected inputs for a given function

The function's input will be mapped to a periodic space, where:
  - if float period is provided, set the period; otherwise use 2*pi
  - if index tuple is provided, only apply the period at the given indices

Examples:
  >>> @periodic()
  ... def twopi(x):
  ...     return x
  ...
  >>> import numpy as np
  >>> twopi([0, np.pi, 2*np.pi, 3*np.pi, 4*np.pi])
  [0.0, 3.141592653589793, 0.0, 3.141592653589793, 0.0]
  >>>
  >>> @periodic(np.pi, index=(0,3,4))
  ... def onepi(x):
  ...     return x
  ...
  >>> onepi([0, np.pi, 2*np.pi, 3*np.pi, 4*np.pi])
  [0.0, 3.141592653589793, 6.283185307179586, 0.0, 0.0]
    """
    from numbers import Integral
    if isinstance(index, Integral): index = (index,)
    index = [index]
    from numpy import asarray, ndarray, ones, zeros, choose, remainder, pi
    if period is None: period = 2*pi
    period = [period]

    def _index(alist=None):
        index[0] = alist

    def _period(period=None):
        period[0] = (2*pi if period is None else period)

    def dec(f):
        def func(x,*args,**kwds):
            if isinstance(x, ndarray): xtype = asarray
            else: xtype = type(x)
            xp = remainder(x, period[0])
            if index[0] is None:
                mask = ones(xp.size, dtype=bool)
            else:
                mask = zeros(xp.size, dtype=bool)
                idx = filter(lambda i:-xp.size<=i<=xp.size-1, index[0])
                try: mask[sorted(idx, key=abs)] = True
                except IndexError: pass
            xp = choose(mask, (x,xp)).reshape(xp.shape)
            return f(xtype(xp), *args, **kwds)
        func.index = _index
        func.period = _period
        return func
    return dec


def boundsconstrain(min, max, **kwds):
    """build a constraint from the bounds (min,max)

Input:
    min: list of floats specifying the lower bounds
    max: list of floats specifying the upper bounds

Additional Input:
    symbolic: bool, if True, use symbolic constraints [default: True]
    clip: bool, if True, clip exterior values to the bounds [default: True]

Notes:
    Prepares a constraint function that imposes the bounds. The intent is
    so that the bounds and other constraints can be imposed concurrently
    (e.g. using `mystic.constraints.and_`). This is slower but more robust
    than applying the bounds sequential with the other constraints (the
    default for a solver).

    For entries where there is no upper bound, use either `inf` or `None`.
    Similarly for entries where there is no lower bound. When no upper
    bound exists for any of the entries, ``max`` should be an iterable
    with all entries of `inf` or `None`. Again, similarly for lower bounds.

    If `symbolic=True`, use symbolic constraints to impose the bounds;
    otherwise use `mystic.constraints.impose_bounds`. Using `clip=False`
    requires `symbolic=False`.
    """
    #NOTE: symbolic_bounds doesn't handle non-iterables (min=None or min=0)
    symbolic = kwds['symbolic'] if 'symbolic' in kwds else True
    clip = kwds['clip'] if 'clip' in kwds else True
    if symbolic and not clip:
        raise NotImplementedError("symbolic must clip to the nearest bound")
    # build the constraint
    if not symbolic: #XXX: much slower than symbolic
        cons = dict((i,j) for (i,j) in enumerate(zip(min, max)))
        cons = impose_bounds(cons, clip=clip)(lambda x: x)
        return cons
    import mystic.symbolic as ms #XXX: randomness due to sympy?
    cons = ms.symbolic_bounds(min, max) #XXX: how clipping with symbolic?
    cons = ms.generate_constraint(ms.generate_solvers(ms.simplify(cons))) #join?
    return cons


#XXX: enable near_integers and has_unique on selected members of x?
#FIXME: the following don't seem to belong in 'mystic.constraints'

from numpy import round, abs
def near_integers(x): # use as a penalty for int programming
    """the sum of all deviations from int values"""
    return abs(x - round(x)).sum()

def has_unique(x): # use as a penalty for unique numbers
    """check for uniqueness of the members of x"""
    return sum(x.count(xi) for xi in x)
   #return len(x) - len(set(x))


##### collapse constraints #####
from mystic.tools import synchronized, connected

#XXX: this is just a less-flexible tools.synchronized... is it useful?
def _impose_as(mask):
    """generate a function, where some input tracks another input

mask should be a dictionary of positional index and tracked index (e.g. {0:1}),
where keys and values should be different integers. However, if a tuple is
provided instead of the tracked index (e.g. {0:(1,2)}), the second member of
the tuple will be used as an additive offset for the tracked index. The mask
will be applied to the input, before the decorated function is called.

functions are expected to take a single argument, a n-dimensional list or array,
where the mask will be applied to the input array.

operations within a single mask are unordered. If a specific ordering of
operations is required, apply multiple masks in the desired order.

Examples:
    >>> @_impose_as({0:1, 2:(3,10)})
    ... def same(x):
    ...     return x
    ...
    >>> same([0,1,2,3,4,5])
    [1, 1, 13, 3, 4, 5]
    >>> same([-1,-2,-3])
    [-2, -2, -3]
    >>> same([-1,-2,-3,-4])
    [-2, -2, 6, -4]

    >>> @_impose_as({1:2})
    ... @_impose_as({0:1})
    ... def doit(x):
    ...     return [i+1 for i in x]
    ...
    >>> doit([0,1,2,3,4,5])
    [3, 3, 3, 4, 5, 6]
    >>> doit([-1,-2,-3])
    [-2, -2, -2]
    >>> doit([-1,-2,-3,-4])
    [-2, -2, -2, -3]
    """
    from numbers import Integral
    offset = lambda i: (i if isinstance(i, Integral) else (i[0], lambda x:(x+i[1])))
    return synchronized(dict((i,offset(j)) for (i,j) in mask.items()))


def impose_as(mask, offset=None):
    """generate a function, where some input tracks another input

mask should be a set of tuples of positional index and tracked index,
where the tuple should contain two different integers. The mask will be
applied to the input, before the decorated function is called.

The offset is applied to the second member of the tuple, and can accumulate.

Examples:
    >>> @impose_as([(0,1),(3,1),(4,5),(5,6),(5,7)])
    ... def same(x):
    ...   return x
    ...
    >>> same([9,8,7,6,5,4,3,2,1])
    [9, 9, 7, 9, 5, 5, 5, 5, 1]
    >>> same([0,1,0,1])
    [0, 0, 0, 0]
    >>> same([-1,-2,-3,-4,-5,-6,-7])
    [-1, -1, -3, -1, -5, -5, -5]

    >>> @impose_as([(0,1),(3,1),(4,5),(5,6),(5,7)], 10)
    ... def doit(x):
    ...   return x
    ...
    >>> doit([9,8,7,6,5,4,3,2,1])
    [9, 19, 7, 9, 5, 15, 25, 25, 1]
    >>> doit([0,1,0,1])
    [0, 10, 0, 0]
    >>> doit([-1,-2,-3,-4,-5,-6])
    [-1, 9, -3, -1, -5, 5]
    >>> doit([-1,-2,-3,-4,-5,-6,-7])
    [-1, 9, -3, -1, -5, 5, 15]
    """
    import copy
    if offset is None: offset = 0
    def dec(f):
        def func(x, *args, **kwds):
            x = copy.copy(x) #XXX: inefficient
            pairs = connected(mask)
            pairs = pairs.items()
            for i,j in pairs:
                for k in j:
                    try: x[k] = x[i]
                    except IndexError: pass
            pairs = list(mask) #XXX: inefficient
            while pairs: # deal with the offset
                indx,trac = zip(*pairs)
                trac = set(trac)
                for i in trac:
                    try: x[i] += offset
                    except IndexError: pass
                indx = trac.intersection(indx)
                pairs = [m for m in pairs if m[0] in indx]
            return f(x, *args, **kwds)
        func.__wrapped__ = f   #XXX: getattr(f, '__wrapped__', f) ?
        func.__doc__ = f.__doc__
        func.mask = mask
        return func
    return dec


def impose_at(index, target=0.0):
    """generate a function, where some input is set to the target

index should be a set of indices to be fixed at the target. The target
can either be a single value (e.g. float), or a list of values.

Examples:
    >>> @impose_at([1,3,4,5,7], -99)
    ... def same(x):
    ...   return x
    ...
    >>> same([1,1,1,1,1,1,1])
    [1, -99, 1, -99, -99, -99, 1]
    >>> same([1,1,1,1])
    [1, -99, 1, -99]
    >>> same([1,1])
    [1, -99]

    >>> @impose_at([1,3,4,5,7], [0,2,4,6])
    ... def doit(x):
    ...   return x
    ...
    >>> doit([1,1,1,1,1,1,1])
    [1, 0, 1, 2, 4, 6, 1]
    >>> doit([1,1,1,1])
    [1, 0, 1, 2]
    >>> doit([1,1])
    [1, 0]
    """
    def dec(f):
        def func(x, *args, **kwds):
            xtype = type(x)
            x = asarray(list(x)) #XXX: faster to use array(x, copy=True) ?
            x[[i for i in index if i < len(x)]] = target
            if not type(x) is xtype: x = xtype(x) #XXX: xtype(x.tolist()) ?
            return f(x, *args, **kwds)
        func.__wrapped__ = f   #XXX: getattr(f, '__wrapped__', f) ?
        func.__doc__ = f.__doc__
        func.target = target
        return func
    return dec


#NOTE: the above is intended to be used thusly... (e.g. produce constraints)
"""
>>> term = Or(ChangeOverGeneration(),CollapseAt(0.0),CollapseAs())
>>> collapses = collapsed(term(solver, True))
>>> _fixed = collapses.get(term[1].__doc__)
>>> _pairs = collapses.get(term[2].__doc__)
>>> 
>>> @impose_as(_pairs)
... @impose_at(_fixed, 0.0)
... def constrain(x):
...   return x
...
>>> 
"""
#NOTE: and for product_measures... (less decorators the better/faster)
"""
>>> stop = Or(ChangeOverGeneration(),CollapsePosition(),CollapseWeight())
>>> collapses = collapsed(stop(solver, True))
>>> _pos = collapses.get(stop[1].__doc__)
>>> _wts = collapses.get(stop[2].__doc__)
>>>
>>> @impose_measure(npts, _pos, _wts)
... def constrain(x):
...   return x
...
>>> 
"""


from mystic.math.discrete import product_measure
#XXX: split to two decorators?
def impose_measure(npts, tracking={}, noweight={}):
    """generate a function, that constrains measure positions and weights

npts is a tuple of the product_measure dimensionality

tracking is a dict of collapses, or a tuple of dicts of collapses.
a tracking collapse is a dict of {measure: {pairs of indices}}, where the
pairs of indices are where the positions will be constrained to have the
same value, and the weight from the second index in the pair will be removed
and added to the weight of the first index

noweight is a dict of collapses, or a tuple of dicts of collapses.
a noweight collapse is a dict of {measure: {indices}), where the
indices are where the measure will be constrained to have zero weight

Examples:
    >>> pos = {0: {(0,1)}, 1:{(0,2)}}
    >>> wts = {0: {1}, 1: {1, 2}}
    >>> npts = (3,3)
    >>> 
    >>> @impose_measure(npts, pos)
    ... def same(x):
    ...   return x
    ...
    >>> same([.5, 0., .5, 2., 4., 6., .25, .5, .25, 6., 4., 2.])
    [0.5, 0.0, 0.5, 2.0, 2.0, 6.0, 0.5, 0.5, 0.0, 5.0, 3.0, 5.0]
    >>> same([1./3, 1./3, 1./3, 1., 2., 3., 1./3, 1./3, 1./3, 1., 2., 3.])
    [0.6666666666666666, 0.0, 0.3333333333333333, 1.3333333333333335, 1.3333333333333335, 3.3333333333333335, 0.6666666666666666, 0.3333333333333333, 0.0, 1.6666666666666667, 2.666666666666667, 1.6666666666666667]
    >>> 
    >>> @impose_measure(npts, {}, wts)
    ... def doit(x):
    ...   return x
    ...
    >>> doit([.5, 0., .5, 2., 4., 6., .25, .5, .25, 6., 4., 2.])
    [0.5, 0.0, 0.5, 2.0, 4.0, 6.0, 1.0, 0.0, 0.0, 4.0, 2.0, 0.0]
    >>> doit([1./3, 1./3, 1./3, 1., 2., 3., 1./3, 1./3, 1./3, 1., 2., 3.])
    [0.5, 0.0, 0.5, 1.0, 2.0, 3.0, 1.0, 0.0, 0.0, 2.0, 3.0, 4.0]
    >>> 
    >>> @impose_measure(npts, pos, wts)
    ... def both(x):
    ...   return x
    ...
    >>> both([1./3, 1./3, 1./3, 1., 2., 3., 1./3, 1./3, 1./3, 1., 2., 3.])
    [0.66666666666666663, 0.0, 0.33333333333333331, 1.3333333333333335, 1.3333333333333335, 3.3333333333333335, 1.0, 0.0, 0.0, 2.0, 3.0, 2.0]
    """
    # takes a dict of collapses, or a tuple of dicts of collapses
    if type(tracking) is dict: tracking = (tracking,)
    if type(noweight) is dict: noweight = (noweight,)
    def dec(f):
        def func(x, *args, **kwds):
            # populate a product measure with params
            c = product_measure()
            c.load(x, npts)
            # apply all collapses
            for clps in tracking:
                for k,v in clps.items():
                    c[k].positions, c[k].weights = \
                      impose_collapse(v, c[k].positions, c[k].weights)
            for clps in noweight:
                for k,v in clps.items():
                    c[k].positions, c[k].weights = \
                      impose_unweighted(v, c[k].positions, c[k].weights, False)
            # convert to params and apply function
            return f(c.flatten(), *args, **kwds)
        func.__wrapped__ = f   #XXX: getattr(f, '__wrapped__', f) ?
        func.__doc__ = f.__doc__
        func.npts = npts
        return func
    return dec


def impose_position(npts, tracking):
    """generate a function, that constrains measure positions

npts is a tuple of the product_measure dimensionality

tracking is a dict of collapses, or a tuple of dicts of collapses.
a tracking collapse is a dict of {measure: {pairs of indices}}, where the
pairs of indices are where the positions will be constrained to have the
same value, and the weight from the second index in the pair will be removed
and added to the weight of the first index

Examples:
    >>> pos = {0: {(0,1)}, 1:{(0,2)}}
    >>> npts = (3,3)
    >>> 
    >>> @impose_position(npts, pos)
    ... def same(x):
    ...   return x
    ...
    >>> same([.5, 0., .5, 2., 4., 6., .25, .5, .25, 6., 4., 2.])
    [0.5, 0.0, 0.5, 2.0, 2.0, 6.0, 0.5, 0.5, 0.0, 5.0, 3.0, 5.0]
    >>> same([1./3, 1./3, 1./3, 1., 2., 3., 1./3, 1./3, 1./3, 1., 2., 3.])
    [0.6666666666666666, 0.0, 0.3333333333333333, 1.3333333333333335, 1.3333333333333335, 3.3333333333333335, 0.6666666666666666, 0.3333333333333333, 0.0, 1.6666666666666667, 2.666666666666667, 1.6666666666666667]
    """
    return impose_measure(npts, tracking, {})


def impose_weight(npts, noweight):
    """generate a function, that constrains measure weights

npts is a tuple of the product_measure dimensionality

noweight is a dict of collapses, or a tuple of dicts of collapses.
a noweight collapse is a dict of {measure: {indices}), where the
indices are where the measure will be constrained to have zero weight

Examples:
    >>> wts = {0: {1}, 1: {1, 2}}
    >>> npts = (3,3)
    >>> 
    >>> @impose_weight(npts, wts)
    ... def doit(x):
    ...   return x
    ...
    >>> doit([.5, 0., .5, 2., 4., 6., .25, .5, .25, 6., 4., 2.])
    [0.5, 0.0, 0.5, 2.0, 4.0, 6.0, 1.0, 0.0, 0.0, 4.0, 2.0, 0.0]
    >>> doit([1./3, 1./3, 1./3, 1., 2., 3., 1./3, 1./3, 1./3, 1., 2., 3.])
    [0.5, 0.0, 0.5, 1.0, 2.0, 3.0, 1.0, 0.0, 0.0, 2.0, 3.0, 4.0]
    """
    return impose_measure(npts, {}, noweight)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2025-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
import numpy as np
from mystic.tools import random_seed
random_seed(17)
from mystic.math import Distribution

size = 10
lb = [0]*4
ub = [10]*4
npts = 20
std = 1
dist = Distribution('numpy.random.normal', 0, std)
X = np.random.random(size=(size,4)) * 10
y = np.random.random(size=size) * 100

from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.set_xlim3d(0,10)
ax.set_ylim3d(0,10)
ax.set_zlim3d(0,10)

from mystic.math.grid import fillpts, samplepts, errorpts
data = np.array(sorted(errorpts(lb, ub, npts, data=X, error=None, mtol=.05, dist=None)))#dist)))
#data = np.array(sorted(fillpts(lb, ub, npts, data=X, dist=None)))#dist)))
#data = np.array(sorted(samplepts(lb, ub, npts, dist=None)))#dist)))

a,b,c,d = data.T
print(a.min(),a.max())
print(b.min(),b.max())
print(c.min(),c.max())
print(d.min(),d.max())

a0,b0,c0,d0 = X.T
ax.scatter(a0,b0,c0,c=d0, cmap=plt.hot(), vmin=0,vmax=10, marker='s', edgecolors='k', s=30)
img = ax.scatter(a,b,c,c=d, cmap=plt.hot(), vmin=0,vmax=10, edgecolors='r', s=30)
fig.colorbar(img)
plt.show()



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2010-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
an example of using an interpolator within a surface object
""" #XXX: use interpolator, plotter, and sampler (instead of Surface)?

from surface import Surface
import time


if __name__ == '__main__':
    start = time.time()
    """
    from mystic.models import griewangk
    from mystic.termination import NormalizedChangeOverGeneration as NCOG

    stop = NCOG(1e-4)
    bounds = 2*[(-9.5,9.5)]

    self = Surface(griewangk, maxpts=1000)

    # self.doit(bounds, stop)
    step=100; scale=False; shift=False; density=9; kwds={}

    if not self.sampler.traj: self.sampler.UseTrajectories()
    # get trajectories
    self.Sample(bounds, stop)
    # get interpolated function
    self.Interpolate(**kwds)
    # check extrema  #XXX: put _min,_max in Interpolate? (downsampled)
    f = lambda x,z: (z,self.surrogate(*x))
    print("min: {}; min@f: {}".format(*f(*self._min())))
    print("max: {}; max@f: {}".format(*f(*self._max())))
    # plot surface
    self.Plot(step=step, scale=scale, shift=shift, density=density)
    """

    # parallel configuration
    try:
        from pathos.helpers import freeze_support, shutdown
        freeze_support()
        from pathos.pools import ProcessPool as Pool
       #from pathos.pools import ThreadPool as Pool
       #from pathos.pools import ParallelPool as Pool
    except ImportError:
        from mystic.pools import SerialPool as Pool
        shutdown = lambda x=None:None

    _map = Pool().map

    # tools
    from mystic.termination import VTR, ChangeOverGeneration as COG
    from mystic.termination import NormalizedChangeOverGeneration as NCOG
    from mystic.monitors import LoggingMonitor, VerboseMonitor, Monitor
    from klepto.archives import dir_archive

    stop = NCOG(1e-4)
    disp = True # print optimization summary
    monitor = True # use LoggingMonitor (uses much less memory)
    archive = False # save an archive
    all = True # use EvalMonitor (instead of StepMonitor only)

    traj = not monitor # save all trajectories internally, if no logs

    # cost function
    from mystic.models import griewangk as model
    from mystic.bounds import Bounds
    ndim = 2  # model dimensionality
    bounds = Bounds(-9.5,9.5,n=ndim) # griewangk

    # the ensemble solvers
    from mystic.solvers import BuckshotSolver as sprayer
    # the local solvers
    from mystic.solvers import PowellDirectionalSolver as seeker

    npts = 25    # number of solvers
    retry = 1    # max consectutive iteration retries without a cache 'miss'
    repeat = 0   # number of times to repeat the search
    tol = 8      # rounding precision
    mem = 1      # cache rounding precision
    size = 0     # max in-memory cache size

    #CUTE: 'configure' monitor and archive if they are desired
    if monitor:
        monitor = LoggingMonitor(1) # montor for all runs
        costmon = LoggingMonitor(1, filename='inv.txt') #XXX: log.txt?
    else:
        monitor = costmon = None
    if archive:
        name = getattr(model,'__name__','model')
        ar_name = '__%s_%sD_cache__' % (name,ndim)
        archive = dir_archive(ar_name, serialized=True, cached=False)
        ar_name = '__%s_%sD_invcache__' % (name,ndim)
        ivcache = dir_archive(ar_name, serialized=True, cached=False)
    else:
        archive = ivcache = None

    from mystic.search import Searcher #XXX: init w/ archive, then UseArchive?
    expts,evals = (None,archive) if all else (archive, None)
    #expts,evals = (archive, None) #XXX: don't override the sample archive
    sampler = Searcher(npts, retry, tol, mem, size, _map, evals, expts, sprayer, seeker, repeat=repeat)
    sampler.Verbose(disp)
    sampler.UseTrajectories(traj)

    ### doit ###
    maxpts = 1000. #10000.
    surface = Surface(model, sampler, maxpts=maxpts, dim=ndim)
    surface.UseMonitor(monitor, costmon)
    surface.UseArchive(archive, ivcache)

    density = 9
    shift = 0
    scale = 0
    step = 200
    args = {
   #'smooth': 0,
    'method': 'thin_plate',
    'extrap': False,
    'arrays': False,
    }
    #surface.doit(bounds, stop, step=step)
   #'multiquadric','inverse','gaussian','linear','cubic','quintic','thin_plate'

    # impose data filter (so X >= 0 and Y >= 0) 
    from mystic.filters import generate_mask, generate_filter
    from mystic.constraints import impose_bounds
    _bounds = impose_bounds((0.0, None))(lambda x:x)
    filter = generate_filter(generate_mask(_bounds, _bounds))
    # from numpy import round as _round
    #############

    # get trajectories
    surface.Sample(bounds, stop, all=all, filter=filter)#, constraints=_round)
    print("TOOK: %s" % (time.time() - start))
#   exit()
    # get interpolated function
    surface.Interpolate(**args)
    # check extrema  #XXX: put _min,_max in Interpolate? (downsampled)
    f = lambda x,z: (z,surface.surrogate(*x))
    print("min: {}; min@f: {}".format(*f(*surface._min())))
    print("max: {}; max@f: {}".format(*f(*surface._max())))
    # shutdown worker pool
    shutdown()
#   print("TOOK: %s" % (time.time() - start))

    # plot interpolated surface
    axes = (0,1)
    vals = () # use remaining minima as the fixed values
    surface.Plot(step=step, scale=scale, shift=shift, density=density, axes=axes, vals=vals)

    """
    try:
        from klepto.archives import file_archive
        archive = file_archive('models.pkl', serialized=True, cached=False)
        archive[model.im_class.__name__.lower()] = surface.surrogate
    except Exception:
        print("serialization failed")
    """
    # some testing of interpolated model
    #import numpy as np
    #actual = np.asarray(surface.z)           # downsample?
    #interp = surface.surrogate(*surface.x.T) # downsample? #NOTE: SLOW
    #print("sum diff squares") #NOTE: is *worse* than with test_searcher.py
    #print("actual and interp: %s" % np.sum((actual - interp)**2))

# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
__doc__ = _doc = """
This is part of Pohlheim's "GEATbx" test suite in [1], with function
definitions drawn from [1], [2], [3], [4], [5], [6], and [7].

References:
    1. Pohlheim, H. "GEATbx: Genetic and Evolutionary Algorithm Toolbox
       for use with MATLAB", Version 3.80, 2006. http://www.geatbx.com/docu
    2. Schwefel, H.-P. "Numerical Optimization of Computer Models",
       John Wiley and Sons, Chichester UK, 1981.
    3. Ackley, D.H. "A Connectionist Machine for Genetic Hillclimbing",
       Kluwer Academic Publishers, Boston MA, 1987.
    4. Michalewicz, Z. "Genetic Algorithms + Data Structures = Evolution
       Programs", Springer-Verlag, Berlin, Heidelberg, New York, 1992.
    5. Branin, F.K. "A Widely Convergent Method for Finding Multiple
       Solutions of Simultaneous Nonlinear Equations", IBM J. Res. Develop.,
       504-522, Sept 1972.
    6. Easom, E.E. "A Survey of Global Optimization Techniques", M. Eng.
       Thesis, U. Louisville, Louisville KY, 1990.
    7. Goldstein, A.A. and Price, I.F. "On Descent from Local Minima",
       Math. Comput., (25) 115, 1971.
"""
#   8. Dixon, L.C.W. and Szego, G.P. "The Optimization Problem: An
#      Introduction", in "Toward Global Optimization II", North Holland,
#      New York, 1978.
__all__ = ['Schwefel', 'HyperEllipsoid', 'Rastrigin', 'DifferentPowers',
           'Ackley', 'Michalewicz', 'Michalewicz', 'Branins', 'Easom',
           'GoldsteinPrice', 'schwefel', 'ellipsoid', 'rastrigin', 'powers',
           'ackley', 'michal', 'branins', 'easom', 'goldstein']

from .abstract_model import AbstractFunction

from math import sin, cos, sqrt, pi, exp

class Schwefel(AbstractFunction):
    __doc__ = \
    """a Schwefel's function generator

Schwefel's function [1,2] has alternating rows of peaks and valleys,
with the global minimum near the edge of the bounded parameter
space.  This funciton can be misleading for optimizers as the next
best local minima are near the other corners of the bounded
parameter space.  The intensity of the peaks and valleys increases
as one moves away from the origin.

The generated function f(x) is identical to function (7)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=2): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Schwefel's function for a list of coeffs

f(x) = \sum_(i=0)^(N-1) -x_i * \sin(\sqrt(\abs(x_i)))

Where abs(x_i) <= 500.

Inspect with mystic_model_plotter using::
    mystic.models.schwefel -b "-500:500:10, -500:500:10" -d

The minimum is f(x)=-(N+1)*418.98288727243374 at x_i=420.9687465 for all i"""
        return sum(-c * sin(sqrt(abs(c))) for c in coeffs)

    minimizers = None #XXX: there are many local minima
    pass


class HyperEllipsoid(AbstractFunction):
    __doc__ = \
    """a Pohlheim's rotated hyper-ellipsoid function generator

Pohlheim's rotated hyper-ellipsoid function [1] is continuous,
convex, and unimodal. The global minimum is located at the
center of the N-dimensional axis parallel hyper-ellipsoid.

The generated function f(x) is identical to function (1b)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=2): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates the rotated hyper-ellipsoid function for a list of coeffs

f(x) = \sum_(i=0)^(N-1) (\sum_(j=0)^(i) x_j)^2

Inspect with mystic_model_plotter using::
    mystic.models.ellipsoid -b "-5:5:.1, -5:5:.1" -d

The minimum is f(x)=0.0 at x_i=0.0 for all i"""
        ncoeffs = range(len(coeffs))
        return sum(sum(coeffs[0:i+1])**2 for i in ncoeffs)

    minimizers = [0.]
    pass


class Rastrigin(AbstractFunction):
    __doc__ = \
    """a Rastrigin's function generator

Rastrigin's function [1] is essentially De Jong's sphere with the
addition of cosine modulation to produce several regularly distributed
local minima. The global minimum is at the origin.

The generated function f(x) is identical to function (6)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=2): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Rastrigin's function for a list of coeffs

f(x) = 10 * N + \sum_(i=0)^(N-1) (x_(i)^2 - 10 * \cos(2 * \pi * x_(i)))

Inspect with mystic_model_plotter using::
    mystic.models.rastrigin -b "-5:5:.1, -5:5:.1" -d

The minimum is f(x)=0.0 at x_i=0.0 for all i"""
        return 10.*len(coeffs) + sum(c*c - 10.*cos(2*pi*c) for c in coeffs)

    minimizers = None #XXX: there are many local minima
    pass


class DifferentPowers(AbstractFunction):
    __doc__ = \
    """a Pohlheim's sum of different powers function generator

Pohlheim's sum of different powers function [1] is unimodal, and
similar to the hyper-ellipsoid and De Jong's sphere. The global
minimum is at the origin, at the center of a broad basin.

The generated function f(x) is identical to function (9)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=2): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates the sum of different powers function for a list of coeffs

f(x) = \sum_(i=0)^(N-1) \abs(x_(i))^(i+2)

Inspect with mystic_model_plotter using::
    mystic.models.powers -b "-5:5:.1, -5:5:.1" -d

The minimum is f(x)=0.0 at x_i=0.0 for all i"""
        ncoeffs = range(len(coeffs))
        return sum(abs(coeffs[i])**(i+2) for i in ncoeffs)

    minimizers = [0.]
    pass


class Ackley(AbstractFunction):
    __doc__ = \
    """an Ackley's path function generator

At a very coarse level, Ackley's path function [1,3] is a slightly
parabolic plane, with a sharp cone-shaped depression at the origin.
The global minimum is found at the origin. There are several local
minima evenly distributed across the function surface, where the
surface modulates similarly to cosine.

The generated function f(x) is identical to function (10)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=2): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Ackley's function for a list of coeffs

f(x) = f_0(x) + f_1(x)

Where:
f_0(x) = -20 * \exp(-0.2 * \sqrt(1/N * \sum_(i=0)^(N-1) x_(i)^(2)))
and:
f_1(x) = -\exp(1/N * \sum_(i=0)^(N-1) cos(2 * \pi * x_(i))) + 20 + \exp(1)

Inspect with mystic_model_plotter using::
    mystic.models.ackley -b "-10:10:.1, -10:10:.1" -d

The minimum is f(x)=0.0 at x_i=0.0 for all i"""
        a=20.; b=0.2; d=2*pi
        n = len(coeffs)
        f0 = -a * exp(-b * sqrt(sum(c*c for c in coeffs)/n))
        f1 = -exp(sum(cos(d*c) for c in coeffs)/n) + a + exp(1)
        return f0 + f1

    minimizers = None #XXX: there are many local minima
    pass


class Michalewicz(AbstractFunction):
    __doc__ = \
    """a Michalewicz's function generator

Michalewicz's function [1,4] in general evaluates to zero. However,
there are long narrow channels that create local minima. At the
intersection of the channels, the function additionally has sharp
dips -- one of which is the global minimum.

The generated function f(x) is identical to function (12)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=5): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Michalewicz's function for a list of coeffs

f(x) = -\sum_(i=0)^(N-1) \sin(x_i) * (\sin((i+1) * (x_i)^(2) / \pi))^(20)

Inspect with mystic_model_plotter using::
    mystic.models.michal -b "0:3.14:.1, 0:3.14:.1, 1.28500168, 1.92305311, 1.72047194" -d

For x=(2.20289811, 1.57078059, 1.28500168, 1.92305311, 1.72047194, ...)[:N]
and c=(-0.801303, -1.0, -0.959092, -0.896699, -1.030564, ...)[:N],
the minimum is f(x)=sum(c) for all x_i=(0,pi)"""
        m = 20 # the larger the number the narrower the channel
        n = range(len(coeffs))
        r = -sum(sin(coeffs[i])*(sin((i+1)*coeffs[i]**2/pi))**m for i in n)
        return r #XXX: constrain x_i=(0,pi) ?

    minimizers = None #XXX: there are many local minima
    pass


class Branins(AbstractFunction):
    __doc__ = \
    """a Branins's rcos function generator

Branins's function [1,5] is very similar to Rosenbrock's saddle function.
However unlike Rosenbrock's saddle, Branins's function has a degenerate
global minimum.

The generated function f(x) is identical to function (13)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=2): # is 2-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Branins's function for a list of coeffs

f(x) = f_0(x) + f_1(x)

Where: 
f_0(x) = a * (x_1 - b * x_(0)^(2) + c * x_0 - d)^2
and
f_1(x) = e * (1 - f) * \cos(x_0) + e
and
a=1, b=5.1/(4*pi^2), c=5/pi, d=6, e=10, f=1/(8*pi)

Inspect with mystic_model_plotter using::
    mystic.models.branins -b "-10:20:.1, -5:25:.1" -d -x 1

The minimum is f(x)=0.397887 at x=((2 +/- (2*i)+1)*pi, 2.275 + 10*i*(i+1)/2)
for all i"""
        x1,x2 = coeffs
        a=1.; b=5.1/(4*pi**2); c=5./pi; d=6.; e=10.; f=1./(8*pi)
        f0 = a * (x2 - b * x1**2 + c * x1 - d)**2
        f1 = e * (1 - f) * cos(x1) + e
        return f0 + f1

    minimizers = None #XXX: there are degenerate global minima
    pass


class Easom(AbstractFunction):
    __doc__ = \
    """a Easom's function generator

Easom's function [1,6] is a unimodal function that exaluates
to zero everywhere except in the region around the global
minimum. The global minimum is at the bottom of a sharp well.

The generated function f(x) is identical to function (14)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=2): # is 2-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Easom's function for a list of coeffs

f(x) = -\cos(x_0) * \cos(x_1) *  \exp(-((x_0-\pi)^2+(x_1-\pi)^2))

Inspect with mystic_model_plotter using::
    mystic.models.easom -b "-5:10:.1, -5:10:.1" -d

The minimum is f(x)=-1.0 at x=(pi,pi)"""
        x0,x1 = coeffs
        return -cos(x0) * cos(x1) *  exp(-((x0-pi)**2+(x1-pi)**2))

    minimizers = [(pi,pi)]
    pass


class GoldsteinPrice(AbstractFunction):
    __doc__ = \
    """a Goldstein-Price's function generator

Goldstein-Price's function [1,7] provides a function with several
peaks surrounding a roughly flat valley. There are a few shallow
scorings across the valley, where the global minimum is found at
the intersection of the deepest of the two scorings. Local minima
occur at other intersections of scorings.

The generated function f(x) is identical to function (15)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=2): # is 2-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Goldstein-Price's function for a list of coeffs

f(x) = (1 + (x_0 + x_1 + 1)^2 * f_0(x)) * (30 + (2*x_0 - 3*x_1)^2 * f_1(x))

Where:
f_0(x) = 19 - 14*x_0 + 3*x_(0)^2 - 14*x_1 + 6*x_(0)*x_(1) + 3*x_(1)^2
and
f_1(x) = 18 - 32*x_0 + 12*x_(0)^2 + 48*x_1 - 36*x_(0)*x_(1) + 27*x_(1)^2

Inspect with mystic_model_plotter using::
    mystic.models.goldstein -b "-5:5:.1, -5:5:.1" -d -x 1

The minimum is f(x)=3.0 at x=(0,-1)"""
        x0,x1 = coeffs
        f0 = 19 - 14*x0 + 3*x0**2 - 14*x1 + 6*x0*x1 + 3*x1**2
        f1 = 18 - 32*x0 + 12*x0**2 + 48*x1 - 36*x0*x1 + 27*x1**2
        return (1 + (x0 + x1 + 1)**2 * f0) * (30 + (2*x0 - 3*x1)**2 * f1)

    minimizers = [(0,-1),(-.6,-.4),(1.8,.2)]
    pass


#class SixHumpCamel(AbstractFunction):
#    __doc__ = \
#    """a Dixon's six-hump camelback function generator
#
#Dixon's six-hump camelback function [1,8] is a saddle
#with two small local minima on the back of the saddle.
#
#The generated function f(x) is identical to function (16)
#of [1], where len(x) >= 0.
#    """ + _doc
#    def __init__(self, ndim=2): # is 2-dimensional
#        AbstractFunction.__init__(self, ndim=ndim)
#        return
#
#    def function(self,coeffs):
#        r"""evaluates Dixon's six-hump camelback function for a list of coeffs
#
#f(x) = f_0(x) + f_1(x)
#
#Where:
#f_0(x) = (4 - 2.1*x_(0)^2 + x_(0)^(4/3)) * x_(0)^2
#and
#f_1(x) = x_(0)*x_(1) + (4*x_(1)^2 - 4) * x_(1)^2
#
#Inspect with mystic_model_plotter using::
#    mystic.models.camelback6 -b "-3:3:.1, -3:3:.1" -d -x 1
#
#The minimum is f(x)=-1.031313 at x=(-0.08834623*j,0.71256506*j) for j = -1,1"""
#        x0,x1 = coeffs
#        f0 = (4 - 2.1*x0**2 + (x0**4)**(1/3.)) * x0**2
#        f1 = x0*x1 + (-4 + 4*x1**2) * x1**2
#        return f0 + f1
#
#    minimizers = None #FIXME: global minimum = -inf
#    pass


# cleanup
del _doc

# prepared instances
schwefel = Schwefel().function
ellipsoid = HyperEllipsoid().function
rastrigin = Rastrigin().function
powers = DifferentPowers().function
ackley = Ackley().function
michal = Michalewicz().function
branins = Branins().function
easom = Easom().function
goldstein = GoldsteinPrice().function
#camelback6 = SixHumpCamel().function


# End of file



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
This module contains classes that aid in constructing cost functions.
Cost function can easily be created by hand; however, mystic also
provides an automated method that allows the dynamic wrapping of 
forward models into cost function objects.

Usage
=====

The basic usage pattern for a cost factory is to generate a cost function
from a set of data points and a corresponding set of evaluation points.
The cost factory requires a "model factory", which is just a generator
of model function instances from a list of coefficients. The following example
uses numpy.poly1d, which provides a factory for generating polynomials. An
expanded version of the following can be found in `mystic.examples.example12`.

    >>> # get a model factory
    >>> import numpy as np
    >>> FunctionFactory = np.poly1d
    >>> 
    >>> # generate some evaluation points
    >>> xpts = 0.1 * np.arange(-50.,51.)
    >>> 
    >>> # we don't have real data, so generate fake data from target and model
    >>> target = [2.,-5.,3.]
    >>> ydata = FunctionFactory(target)(xpts)
    >>> noise = np.random.normal(0,1,size=len(ydata))
    >>> ydata = ydata + noise
    >>> 
    >>> # get a cost factory
    >>> from mystic.forward_model import CostFactory
    >>> C = CostFactory()
    >>> 
    >>> # generate a cost function for the model factory
    >>> metric = lambda x: np.sum(x*x)
    >>> C.addModel(FunctionFactory, inputs=len(target))
    >>> cost = C.getCostFunction(evalpts=xpts, observations=ydata,
    ...                                        sigma=1.0, metric=metric)
    >>>
    >>> # pass the cost function to the optimizer
    >>> initial_guess = [1.,-2.,1.]
    >>> solution = fmin_powell(cost, initial_guess)
    >>> print(solution)
    [ 2.00495233 -5.0126248   2.72873734]


In general, a user will be required to write their own model factory.
See the examples contained in `mystic.models` for more information.

The CostFactory can be used to couple models together into a single cost
function. For an example, see `mystic.examples.forward_model`.

"""

from mystic.filters import identity, component
from mystic.filters import null_check

#from inspect import getargspec
from numpy import pi, sqrt, array, mgrid, random, real, conjugate, arange, sum

__all__ = ['CostFactory']

class CostFactory(object):
    """
A cost function generator.
    """
    def __init__(self):
        """
CostFactory builds a list of forward model factories, and maintains a list
of associated model names and number of inputs. Can be used to combine several
models into a single cost function.

Takes no initial inputs.
        """
        self._names = []
        self._forwardFactories = []
        self._inputs = []
        self._inputFilters = {}
        self._outputFilters = []
        self._inputCheckers = []
        pass

    def addModel(self, model, inputs, name=None, outputFilter=identity, inputChecker=null_check):
        """
Adds a forward model factory to the cost factory.

Inputs:
    model   -- a callable function factory object
    inputs  -- number of input arguments to model
    name    -- a string representing the model name

Example:
    >>> import numpy as np
    >>> C = CostFactory()
    >>> C.addModel(np.poly, inputs=3)
        """
        if name is None:
            import dill
            name = dill.source.getname(model)
        if name is None:
            for i in range(len(self._names)+1):
                name = 'model'+str(i)
                if name not in self._names: break
        elif name in self._names:
            print("Model [%s] already in database." % name)
            raise AssertionError
        self._names.append(name)
        self._forwardFactories.append(model)
        self._inputs.append(inputs)
        self._outputFilters.append(outputFilter)
        self._inputCheckers.append(inputChecker)

    #XXX: addModelNew is a work in progress...
    '''
    def addModelNew(self, model, name, outputFilter=identity, inputChecker=null_check):
        """
Adds a forward model factory to the cost factory.
The number of inputs is determined with inspect.getargspec.

Inputs:
    model   -- a callable function factory object
    name    -- a string representing the model name
        """
        #NOTE: better to replace "old-style" addModel above?
        if name in self._names:
            print("Model [%s] already in database." % name)
            raise AssertionError
        self._names.append(name)
        self._forwardFactories.append(model)
        inputs = getargspec(model)[0]
        self._inputs.append(len(inputs))
        self._outputFilters.append(outputFilter)
        self._inputCheckers.append(inputChecker)
    '''

    def getForwardEvaluator(self, evalpts):
        """
Get a model factory that allows simultaneous evaluation of all forward models
for the same set of evaluation points.

Inputs:
    evalpts -- a list of evaluation points

Example:
    >>> import numpy as np
    >>> C = CostFactory()
    >>> C.addModel(np.poly, inputs=3)
    >>> F = C.getForwardEvaluator([1,2,3,4,5])
    >>> F([1,0,0])
    [array([ 1,  4,  9, 16, 25])]
    >>> F([0,1,0])
    [array([1, 2, 3, 4, 5])]
        """
        #NOTE: does NOT go through inputChecker
        def _(params):
            out = []
            ind = 0
            for F, n, ofilt in zip(self._forwardFactories, self._inputs, self._outputFilters):
                m = F(params[ind:ind+n])
                out.append(ofilt(m(evalpts)))
                ind = ind+n
            return out
        return _

    def getVectorCostFunction(self, evalpts, observations):
        """
Get a vector cost function that allows simultaneous evaluation of all
forward models for the same set of evaluation points and observation points.

Inputs:
    evalpts -- a list of evaluation points
    observations -- a list of data points

The vector cost metric is hard-wired to be the sum of the difference of
getForwardEvaluator(evalpts) and the observations.

NOTE: Input parameters do NOT go through filters registered as inputCheckers.

Example:
    >>> import numpy as np
    >>> C = CostFactory()
    >>> C.addModel(np.poly, inputs=3)
    >>> x = np.array([-2., -1., 0., 1., 2.])
    >>> y = np.array([-4., -2., 0., 2., 4.])
    >>> F = C.getVectorCostFunction(x, y)
    >>> F([1,0,0])
    0.0
    >>> F([2,0,0])
    10.0
        """
        def _(params):
            forward = self.getForwardEvaluator(evalpts)
            return sum(forward(params) - observations)
        return _

    def getCostFunction(self, evalpts, observations, sigma=None, metric=lambda x: sum(x*x)):
        """
Get a cost function that allows simultaneous evaluation of all forward models
for the same set of evaluation points and observation points.

Inputs:
    evalpts -- a list of evaluation points
    observations -- a list of data points
    sigma   -- a scaling factor applied to the raw cost
    metric  -- the cost metric object

The cost metric should be a function of one parameter (possibly an array)
that returns a scalar. The default is L2. When called, the "misfit" will
be passed in.

NOTE: Input parameters WILL go through filters registered as inputCheckers.

Example:
    >>> import numpy as np
    >>> C = CostFactory()
    >>> C.addModel(np.poly, inputs=3)
    >>> x = np.array([-2., -1., 0., 1., 2.])
    >>> y = np.array([-4., -2., 0., 2., 4.])
    >>> F = C.getCostFunction(x, y, metric=lambda x: np.sum(x))
    >>> F([1,0,0])
    0.0
    >>> F([2,0,0])
    10.0
    >>> F = C.getCostFunction(x, y)
    >>> F([2,0,0])
    34.0
        """
        #XXX: better interface for sigma?
        def _(params):
            ind = 0
            for F, n, ofilt, icheck in zip(self._forwardFactories, self._inputs, \
                                       self._outputFilters, self._inputCheckers):
                # check input  #XXX: is this worthwile to do?
                my_params = params[ind:ind+n]
                checkQ = icheck(my_params, evalpts)
                if checkQ is not None:
                    # some parameters are out of range... returns "cost"
                    return checkQ

                Gm = F(params[ind:ind+n])
                if ind == 0:
                    x = ofilt(Gm(evalpts)) 
                else:
                    x = x + ofilt(Gm(evalpts)) 
                ind = ind+n
            if sigma is None:
                x = x - observations
            else:
                x = (x - observations) / sigma
            #return sum(real((conjugate(x)*x)))
            #return sum(x*x) 
            return metric(x)
        return _

    def getCostFunctionSlow(self, evalpts, observations):
        """Get a cost function that allows simultaneous evaluation of all
forward models for the same set of evaluation points and observation points.

Args:
    evalpts (list(float)): a list of evaluation points (i.e. input).
    observations (list(float)): a list of data points (i.e. output).

Notes:
    The cost metric is hard-wired to be the sum of the real part of ``|x|^2``,
    where x is the VectorCostFunction for a given set of parameters.

    Input parameters do NOT go through filters registered as inputCheckers.

Examples:
    >>> import numpy as np
    >>> C = CostFactory()
    >>> C.addModel(np.poly, inputs=3)
    >>> x = np.array([-2., -1., 0., 1., 2.])
    >>> y = np.array([-4., -2., 0., 2., 4.])
    >>> F = C.getCostFunctionSlow(x, y)
    >>> F([1,0,0])
    0.0
    >>> F([2,0,0])
    100.0
        """
        #XXX: update interface to allow metric?
        def _(params):
            v = self.getVectorCostFunction(evalpts, observations)
            x = v(params)
            return sum(real((conjugate(x)*x)))
        return _

    def getParameterList(self):
        """
Get a 'pretty' listing of the input parameters and corresponding models.
        """
        inputList = []
        for name, n in zip(self._names, self._inputs):
            inputList += ['%s.x%d' % (name, i) for i in range(n)]
        return inputList

    def getRandomParams(self):
        import random
        return array([random.random() for i in range(sum(self._inputs))])
    
    def __repr__(self):
        return "Input Parameter List: %s " % self.getParameterList()
            

if __name__=='__main__':
    pass


# End of file



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

def objective(x):
    from numpy import prod, sqrt
    n = len(x)
    return -sqrt(n)*n * prod(x)

def bounds(len=3):
    return [(0.0,1.0)]*len

# with penalty='penalty' applied, solution is:
def xs(len=3):
    from math import sqrt
    return [1./sqrt(len)]*len
def ys(len=3):
    return objective(xs(len))

"""
for len(x) == 3, x* = 0.57735027 for all xi, y* = -1.0
for len(x) == 10, x* = 0.31622777 for all xi. y* = -0.00031623
for len(x) == 20, x* = 0.22360680 for all xi, y* = -8.73464054e-12
"""

from mystic.symbolic import generate_constraint, generate_solvers, simplify
from mystic.symbolic import generate_penalty, generate_conditions

#sum([x0**2, x1**2, x2**2]) - 1.0 = 0.0
def equations(len=3):
    eqn = "\nsum(["
    for i in range(len):
        eqn += 'x%s**2, ' % str(i)
    return eqn[:-2]+"]) - 1.0 = 0.0\n"

def cf(len=3):
    return generate_constraint(generate_solvers(simplify(equations(len))))
def pf(len=3):
    return generate_penalty(generate_conditions(equations(len)))



if __name__ == '__main__':
    x = xs(10)
    y = ys(len(x))
    bounds = bounds(len(x))
    cf = cf(len(x))

    from mystic.solvers import diffev2
    from mystic.math import almostEqual

    result = diffev2(objective, x0=bounds, bounds=bounds, constraints=cf, npop=40, gtol=500, disp=False, full_output=True)

    assert almostEqual(result[0], x, tol=1e-2)
    assert almostEqual(result[1], y, tol=1e-2)



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.constraints import *
from mystic.penalty import quadratic_equality
from mystic.coupler import inner
from mystic.math import almostEqual
from mystic.tools import random_seed
random_seed(669)

def test_penalize():

  from mystic.math.measures import mean, spread
  def mean_constraint(x, target):
    return mean(x) - target

  def range_constraint(x, target):
    return spread(x) - target

  @quadratic_equality(condition=range_constraint, kwds={'target':5.0})
  @quadratic_equality(condition=mean_constraint, kwds={'target':5.0})
  def penalty(x):
    return 0.0

  def cost(x):
    return abs(sum(x) - 5.0)

  from mystic.solvers import fmin
  from numpy import array
  x = array([1,2,3,4,5])
  y = fmin(cost, x, penalty=penalty, disp=False)

  assert round(mean(y)) == 5.0
  assert round(spread(y)) == 5.0
  assert round(cost(y)) == 4*(5.0)


def test_solve():

  from mystic.math.measures import mean
  def mean_constraint(x, target):
    return mean(x) - target

  def parameter_constraint(x):
    return x[-1] - x[0]

  @quadratic_equality(condition=mean_constraint, kwds={'target':5.0})
  @quadratic_equality(condition=parameter_constraint)
  def penalty(x):
    return 0.0

  x = solve(penalty, guess=[2,3,1])
 #x = solve(penalty, solver='buckshot')

  assert round(mean_constraint(x, 5.0)) == 0.0
  assert round(parameter_constraint(x)) == 0.0
  assert issolution(penalty, x, tol=0.002)


def test_solve_constraint():

  from mystic.math.measures import mean
  @with_mean(1.0)
  def constraint(x):
    x[-1] = x[0]
    return x

  x = solve(constraint, guess=[2,3,1])
 #x = solve(constraint, solver='buckshot')

  assert almostEqual(mean(x), 1.0, tol=1e-15)
  assert x[-1] == x[0]
  assert issolution(constraint, x)


def test_as_constraint():

  from mystic.math.measures import mean, spread
  def mean_constraint(x, target):
    return mean(x) - target

  def range_constraint(x, target):
    return spread(x) - target

  @quadratic_equality(condition=range_constraint, kwds={'target':5.0})
  @quadratic_equality(condition=mean_constraint, kwds={'target':5.0})
  def penalty(x):
    return 0.0

  ndim = 3
  constraints = as_constraint(penalty)#, solver='fmin')
  #XXX: this is expensive to evaluate, as there are nested optimizations

  from numpy import arange
  x = arange(ndim)
  _x = constraints(x)
  
  assert round(mean(_x)) == 5.0
  assert round(spread(_x)) == 5.0
  assert round(penalty(_x)) == 0.0

  def cost(x):
    return abs(sum(x) - 5.0)

  npop = ndim*3
  from mystic.solvers import diffev
  y = diffev(cost, x, npop, constraints=constraints, disp=False, gtol=10)

  assert round(mean(y)) == 5.0
  assert round(spread(y)) == 5.0
  assert round(cost(y)) == 5.0*(ndim-1)


def test_as_penalty():

  from mystic.math.measures import mean, spread
  @with_spread(5.0)
  @with_mean(5.0)
  def constraint(x):
    return x

  penalty = as_penalty(constraint)

  from numpy import array
  x = array([1,2,3,4,5])
  
  def cost(x):
    return abs(sum(x) - 5.0)

  from mystic.solvers import fmin
  y = fmin(cost, x, penalty=penalty, disp=False)

  assert round(mean(y)) == 5.0
  assert round(spread(y)) == 5.0
  assert round(cost(y)) == 4*(5.0)


def test_with_penalty():

  from mystic.math.measures import mean, spread
  @with_penalty(quadratic_equality, kwds={'target':5.0})
  def penalty(x, target):
    return mean(x) - target

  def cost(x):
    return abs(sum(x) - 5.0)

  from mystic.solvers import fmin
  from numpy import array
  x = array([1,2,3,4,5])
  y = fmin(cost, x, penalty=penalty, disp=False)

  assert round(mean(y)) == 5.0
  assert round(cost(y)) == 4*(5.0)


def test_with_mean():

  from mystic.math.measures import mean, impose_mean

  @with_mean(5.0)
  def mean_of_squared(x):
    return [i**2 for i in x]

  from numpy import array
  x = array([1,2,3,4,5])
  y = impose_mean(5, [i**2 for i in x])
  assert mean(y) == 5.0
  assert mean_of_squared(x) == y


def test_with_mean_spread():

  from mystic.math.measures import mean, spread, impose_mean, impose_spread

  @with_spread(50.0)
  @with_mean(5.0)
  def constrained_squared(x):
    return [i**2 for i in x]

  from numpy import array
  x = array([1,2,3,4,5])
  y = impose_spread(50.0, impose_mean(5.0,[i**2 for i in x]))
  assert almostEqual(mean(y), 5.0, tol=1e-15)
  assert almostEqual(spread(y), 50.0, tol=1e-15)
  assert constrained_squared(x) == y


def test_constrained_solve():

  from mystic.math.measures import mean, spread
  @with_spread(5.0)
  @with_mean(5.0)
  def constraints(x):
    return x

  def cost(x):
    return abs(sum(x) - 5.0)

  from mystic.solvers import fmin_powell
  from numpy import array
  x = array([1,2,3,4,5])
  y = fmin_powell(cost, x, constraints=constraints, disp=False)

  assert almostEqual(mean(y), 5.0, tol=1e-15)
  assert almostEqual(spread(y), 5.0, tol=1e-15)
  assert almostEqual(cost(y), 4*(5.0), tol=1e-6)


def test_with_constraint():

  from mystic.math.measures import mean, impose_mean

  @with_constraint(inner, kwds={'target':5.0})
  def mean_of_squared(x, target):
    return impose_mean(target, [i**2 for i in x])

  from numpy import array
  x = array([1,2,3,4,5])
  y = impose_mean(5, [i**2 for i in x])
  assert mean(y) == 5.0
  assert mean_of_squared(x) == y


def test_discrete():

  @discrete([1.0, 3.5, 5.5, 7.0])
  def discrete_squared(x):
    return x**2

  from numpy import asarray
  assert discrete_squared(5.6) == 5.5**2
  assert all(discrete_squared(asarray([1, 3])) == asarray([1.0, 3.5])**2)
  discrete_squared.samples([1.0, 7.0])
  assert discrete_squared(5.6) == 7.0**2
  discrete_squared.index([0, -1])
  assert all(discrete_squared(asarray([0, 3, 6])) == asarray([1.0, 3.0, 7.0])**2)


def test_sorting():

  x = [1.0, -3.5, -5.5, 7.0, 9.0]
  negative = lambda x: [-i for i in x]
  assert sorting()(negative)(x) == negative(sorted(x))
  assert sorting(outer=True)(negative)(x) == sorted(negative(x))
  assert sorting(ascending=False)(negative)(x) == negative(sorted(x, reverse=True))
  assert sorting(ascending=False, outer=True)(negative)(x) == sorted(negative(x), reverse=True)

  assert sorting()(sum)(x) == sum(x)
  assert sorting(outer=True)(sum)(x) == sum(x)
  assert sorting(ascending=False)(sum)(x) == sum(x)
  assert sorting(ascending=False, outer=True)(sum)(x) == sum(x)

  assert monotonic()(negative)(x) == monotonic(ascending=False, outer=True)(negative)(x)
  assert monotonic(outer=True)(negative)(x) == monotonic(ascending=False)(negative)(x)

  assert sorting(index=0)(negative)(x) == sorting(index=[0])(negative)(x)
  assert sorting(index=(0,1,2))(negative)(x) == sorting(index=(-4,-5,2))(negative)(x)
  assert monotonic(index=0)(negative)(x) == monotonic(index=[0])(negative)(x)
  assert monotonic(index=(0,1,2))(negative)(x) == monotonic(index=(-4,-5,2))(negative)(x)

  from numpy import maximum as max, minimum as min
  assert monotonic()(sum)(x) == sum(max.accumulate(x))
  assert monotonic(outer=True)(sum)(x) == sum(x)
  assert monotonic(ascending=False)(sum)(x) == sum(min.accumulate(x))
  assert monotonic(ascending=False, outer=True)(sum)(x) == sum(x)


def test_unique():
  x = [1, 2, 3, 1, 2, 10]
  y = unique(x)
  z = set(y).difference(x)
  assert all(y.count(i) == 1 for i in y)
  assert all(min(x) <= i < max(x) for i in z)
  assert all(isinstance(i, int) for i in z)
  y = unique(x, float)
  z = set(y).difference(x)
  assert all(y.count(i) == 1 for i in y)
  assert all(min(x) <= i < max(x) for i in z)
  assert all(isinstance(i, float) for i in z)
  y = unique(x, range(11))
  z = set(y).difference(x)
  assert all(y.count(i) == 1 for i in y)
  assert all(0 <= i < 11 for i in z)
  assert all(isinstance(i, int) for i in z)
  y = unique(x, {'min':0, 'max':11})
  z = set(y).difference(x)
  assert all(y.count(i) == 1 for i in y)
  assert all(min(x) <= i < max(x) for i in z)
  assert all(isinstance(i, float) for i in z)
  y = unique(x, {'min':0, 'max':11, 'type':int})
  z = set(y).difference(x)
  assert all(y.count(i) == 1 for i in y)
  assert all(min(x) <= i < max(x) for i in z)
  assert all(isinstance(i, int) for i in z)
  x = [1, 2, 3, 1.0, 2, 10]
  y = unique(x, int)
  z = set(y).difference(x)
  assert all(y.count(i) == 1 for i in y)
  assert all(min(x) <= i < max(x) for i in z)
  assert all(isinstance(i, int) for i in z)
  y = unique(x)
  z = set(y).difference(x)
  assert all(y.count(i) == 1 for i in y)
  assert all(min(x) <= i < max(x) for i in z)
  assert all(isinstance(i, float) for i in z)
  x = [1, 2, 3, 1.5, 2, 10]
  try:
      unique(x, int)
      assert False
  except ValueError:
      assert True
  y = unique(x)
  z = set(y).difference(x)
  assert all(y.count(i) == 1 for i in y)
  assert all(min(x) <= i < max(x) for i in z)
  assert all(isinstance(i, float) for i in z)
  x = [1, 2, 3, 1, 2, 13]
  try:
      unique(x, range(11))
      assert False
  except ValueError:
      assert True
  try:
      unique([1,2,3,1,2,4], int)
      assert False
  except ValueError:
      assert True


if __name__ == '__main__':
  test_penalize()
  test_solve()
  test_solve_constraint()
  test_as_constraint()
  test_as_penalty()
  test_with_penalty()
  test_with_mean()
  test_with_mean_spread()
  test_constrained_solve()
  test_with_constraint()
  test_discrete()
  test_sorting()
  test_unique()


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2022-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
create a standard file archive that captures function input and output, and retains calling order
'''
# generate a function that writes to the archive, where it:
#  - uses function input as the key, and function output as the value
#  - is ordered by function call
import mystic.cache as mc
a = mc.archive.read('cache.db', type=mc.archive.file_archive)
from toys import function5
objective = mc.cached(archive=a)(function5)

# set some bounds
from mystic.bounds import Bounds
bounds = Bounds([0,1,0,0,0],[1,10,10,10,10])

# optimize, using a evaluation monitor
import mystic as my
mon = my.monitors.Monitor()
my.solvers.fmin(objective, x0=bounds.xlb, bounds=bounds, evalmon=mon, full_output=True, disp=False)

# archived rv and cost should match evaluation monitor, regardless of size
# (to look up the last entry, we need to use keys(), so we just check all keys)
# file archives preserve order, but we convert to nested lists to compare
xs,ys = list(list(k) for k in a.keys()), list(a.values())
assert xs == mon.x
assert ys == mon.y

# remove the archive
import os
os.remove('cache.db')


'''
** alternate **
create a limited-size customized directory archive that captures function input and output, and retains calling order
'''
# generate a global counter
from mystic import Counter
counter = Counter(0)
run = lambda : counter.count()

# generate a function that writes to the archive, where it:
#  - uses run number as the key
#  - uses the given kwds as the values
#  - any kwds input to the cached function is ignored in the keys
#  - has a maximum archive size of 500
import mystic.cache as mc
@mc.cached(type=mc.lru_cache, archive='cache', maxsize=500, ignore='**')
def archive(run, **kwds): # key: run[int]
    return kwds # value: kwds[dict]

# embed writing to the archive within the objective
def objective(rv):
    from toys import function5
    cost = function5(rv)
    archive([run()], rv=rv, cost=cost)
    return cost

# set some bounds
from mystic.bounds import Bounds
bounds = Bounds([0,1,0,0,0],[1,10,10,10,10])

# optimize, using a evaluation monitor
import mystic as my
mon = my.monitors.Monitor()
my.solvers.fmin(objective, x0=bounds.xlb, bounds=bounds, evalmon=mon, full_output=True, disp=False)

# read the archive
a = mc.archive.read('cache')

# archived rv and cost should match evaluation monitor
# (for up to 500 function evaluations -- the max size of the archive)
# hence, we will just check the last entry 
last = a[len(a)-1]
assert last['rv'] == mon.x[-1]
assert last['cost'] == mon.y[-1]

# directory archives do not preserve order, so we need to sort by key
xs,ys = zip(*[(a[i]['rv'],a[i]['cost']) for i in sorted(a.keys())])
assert list(xs) == mon.x
assert list(ys) == mon.y

# remove the archive
import shutil
shutil.rmtree('cache')



#!/usr/bin/env python
#
# Problem definition:
# Example in reference documentation for cvxopt
# http://cvxopt.org/examples/tutorial/qp.html
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from cvxqp import objective, bounds, xs, ys
    
from mystic.penalty import quadratic_equality
from mystic.constraints import with_penalty

@with_penalty(quadratic_equality, k=1e4)
def penalty(x): # == 0.0
    return x[1] + x[0] - 1.0


if __name__ == '__main__':

  from mystic.solvers import diffev2, fmin_powell
  from mystic.math import almostEqual

  result = diffev2(objective, x0=bounds, bounds=bounds, penalty=penalty, npop=40, disp=False, full_output=True)
  assert almostEqual(result[0], xs, rel=2e-2)
  assert almostEqual(result[1], ys, rel=2e-2)

  result = fmin_powell(objective, x0=[0.0,0.0], bounds=bounds, penalty=penalty, disp=False, full_output=True)
  assert almostEqual(result[0], xs, rel=2e-2)
  assert almostEqual(result[1], ys, rel=2e-2)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
misc user-defined items (solver configuration, moment constraints)
"""
from mystic.solvers import DifferentialEvolutionSolver2
from mystic.monitors import VerboseMonitor, Monitor
from mystic.termination import ChangeOverGeneration as COG
from mystic.bounds import Bounds, MeasureBounds

# kwds for solver
opts = dict(termination=COG(1e-6, 40), CrossProbability=0.9, ScalingFactor=0.9)
param = dict(solver=DifferentialEvolutionSolver2,
             npop=10, #XXX:npop
             maxiter=1000,
             maxfun=1e+6,
             x0=None, # use RandomInitialPoints
             nested=None, # don't use SetNested
             map=None, # don't use SetMapper
             stepmon=VerboseMonitor(1, label='output'), # monitor config
             #evalmon=Monitor(), # monitor config (re-initialized in solve)
             # kwds to pass directly to Solve(objective, **opt)
             opts=opts,
            )


from mystic.constraints import and_, integers
from mystic.coupler import outer, additive
from ouq_misc import (flatten, unflatten, normalize_moments, constrained, check,
                      constrain_moments, constrain_expected, constrained_out)

# lower and upper bound for parameters and weights
xlb = (20.0,0.0,2.1)
xub = (150.0,30.0,2.8)
wlb = (0,1,1)
wub = (1,1,1)
# number of Dirac masses to use for each parameter
npts = (2,1,1) #NOTE: rv = (w0,w0,x0,x0,w1,x1,w2,x2)
index = (5,)       #NOTE: rv[5] -> x1
# moments and uncertainty in first parameter
a_ave = None
a_var = None
a_ave_err = None
a_var_err = None
# moments and uncertainty in second parameter
b_ave = None
b_var = None
b_ave_err = None
b_var_err = None
# moments and uncertainty in output
o_ave = 6.5
o_var = None
o_ave_err = 1.0
o_var_err = None


# build a model representing 'truth'
from ouq_models import WrapModel
from surrogate import marc_surr as toy; nx = 3; ny = None; Ns = None
nargs = dict(nx=nx, ny=ny, rnd=(True if Ns else False))
model = WrapModel('model', toy, **nargs)

# set the bounds
bnd = MeasureBounds(xlb, xub, n=npts, wlb=wlb, wub=wub)

# constrain parameters at given index(es) to be ints
integer_indices = integers(ints=float, index=index)(lambda rv: rv)

## moment-based constraints ##
normcon = normalize_moments()
###momcons = constrain_moments(a_ave, a_var, a_ave_err, a_var_err)
###is_cons = constrained(a_ave, a_var, a_ave_err, a_var_err)
#momcon0 = constrain_moments(a_ave, a_var, a_ave_err, a_var_err, idx=0)
#momcon1 = constrain_moments(b_ave, b_var, b_ave_err, b_var_err, idx=1)
#is_con0 = constrained(a_ave, a_var, a_ave_err, a_var_err, idx=0)
#is_con1 = constrained(b_ave, b_var, b_ave_err, b_var_err, idx=1)
#is_cons = lambda c: bool(additive(is_con0)(is_con1)(c))
#momcons = constrain_expected(model, o_ave, o_var, o_ave_err, o_var_err, bnd, constraints=normcon)
momcons = constrain_expected(model, o_ave, o_var, o_ave_err, o_var_err, bnd)
is_cons = constrained_out(model, o_ave, o_var, o_ave_err, o_var_err)

## index-based constraints ##
# impose constraints sequentially (faster, but assumes are decoupled)
#scons = outer(integer_indices)(flatten(npts)(outer(momcons)(normcon)))
#scons = flatten(npts)(outer(momcon1)(outer(momcon0)(normcon)))
scons = flatten(npts)(outer(momcons)(normcon))
#scons = flatten(npts)(momcons)

# impose constraints concurrently (slower, but safer)
#ccons = and_(flatten(npts)(normcon), flatten(npts)(momcons), integer_indices)
#ccons = and_(flatten(npts)(normcon), flatten(npts)(momcon0), flatten(npts)(momcon1))
ccons = and_(flatten(npts)(normcon), flatten(npts)(momcons))
#ccons = scons

# check parameters (instead of measures)
iscon = check(npts)(is_cons)



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2018-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Maximization with a boolean variable and constraints.

Maximize:
    sum_{i=1}^{n-1} sum_{j=i+1}^{n} w_{ij} x_{i} x_{j}

where:
    sum_{i=1}^{n} x_{i} < b
    x_{i} in {0,1}
"""
from mystic.solvers import diffev2
from mystic.monitors import VerboseMonitor
from mystic.constraints import impose_sum, discrete, and_
import numpy as np

N = 10
b = 5
from mystic.bounds import Bounds
bounds = Bounds(0,1,n=N)

def objective(x, w):
    s = 0
    for i in range(len(x)-1):
        for j in range(i, len(x)):
            s += w[i,j] * x[i] * x[j]
    return s


w = np.ones((N,N)) #XXX: replace with actual values of wij

cost = lambda x: -objective(x, w)

c = and_(lambda x: impose_sum(b, x), discrete([0,1])(lambda x:x))

mon = VerboseMonitor(10)
solution = diffev2(cost,bounds,constraints=c,bounds=bounds,itermon=mon,gtol=50, maxiter=5000, maxfun=50000, npop=10)
print(solution)




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
OUQ classes for calculating bounds on statistical quantities
"""
from mystic.math.discrete import product_measure
from ouq import BaseOUQ


class MeanValue(BaseOUQ):
    def objective(self, rv, axis=None, idx=0): #FIXME: idx != None, is fixed
        """calculate mean value of input, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)
        idx: int, the index of input to calculate (all, by default) #FIXME

    Returns:
        the mean value for the specified index

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_expect if samples, else expect
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv): #FIXME: set model,samples in constraints
            return self._invalid
        # get mean value
        if axis is None and self.axes is not None:
            model = (lambda x: self.model(x, axis=i) for i in range(self.axes))
            if self.samples is None:
                return NotImplemented #FIXME: set model,samples in constraints
            # else use sampled support
            return NotImplemented #FIXME: set model,samples in constraints
        # else, get mean value for the given axis
        if axis is None:
            model = lambda x: self.model(x)
        else:
            model = lambda x: self.model(x, axis=axis)
        if self.samples is None:
            if idx is None:
                res = (c[i].mean for i in self.nx)
                return tuple(res)
            return c[idx].mean #FIXME: set model,samples in constraints
        return NotImplemented #FIXME: set model,samples in constraints


class Negativity(BaseOUQ):

    def objective(self, rv, axis=None, iter=True):
        """calculate expected negativity for model, under uncertainty

    Input:
        rv: list of input parameters
        axis: int, the index of output to calculate (all, by default)
        iter: bool, if True, calculate per axis, else calculate for all axes

    Returns:
        the expected negativity for the specified axis (or all axes)

    NOTE:
        respects constraints on input parameters and product measure

    NOTE:
        for product_measure, use sampled_pof(func) if samples, else pof(func)
        where func = lambda x: model(x) >= 0
        """
        # check constraints
        c = product_measure().load(rv, self.npts)
        if not self.cvalid(c) or not self.xvalid(rv):
            if axis is None and self.axes is not None:
                return (self._invalid,) * (self.axes or 1) #XXX:?
            return self._invalid
        # get expected negativity
        if iter and axis is None and self.axes is not None:
            model = (lambda x: (self.model(x,axis=i) >= 0) for i in range(self.axes))
            if self.samples is None:
                return tuple(c.pof(m) for m in model)
            # else use sampled support
            return tuple(c.sampled_pof(m, self.samples, map=self.map) for m in model)
        # else, get expected negativity for the given axis
        if axis is None:
            model = lambda x: (self.model(x) >= 0)
        else:
            model = lambda x: (self.model(x, axis=axis) >= 0)
        if self.samples is None:
            return c.pof(model)
        return c.sampled_pof(model, self.samples, map=self.map)




#!/usr/bin/env python
#
# Problem definition and original response:
# https://stackoverflow.com/q/67560280/2379433
# https://stackoverflow.com/a/67571398/2379433
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2021-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Objective:
    MIN (1500 * x0) + (625 * x1) + (100 * x2)

Constraints:
    schedule(x0,x1,x2) >= 480
    x2 = round(log(x0)**2)
    x1 > x0
    x1 > x2
    x0, x1, x2 are integers

Bounds:
    (5 <= x0 <= 50), (5 <= x1 <= 100), (1 <= x2 <= 20)
"""
import mystic as my
import mystic.symbolic as ms
import numpy as np
 
# define the cost function
def objective(x):
    x0,x1,x2 = x
    return (1500 * x0) + (625 * x1) + (100 * x2)

# define the bounds
from mystic.bounds import Bounds
bounds = Bounds([5,5,1],[50,100,20])

# define symbolic constraints
eqns = '''
x1 > x0
x2 < x1
'''

# generate constraint operator from symbolic constraints
and_ = my.constraints.and_
cons = ms.generate_constraint(ms.generate_solvers(ms.simplify(eqns)), join=and_)

''' #NOTE: cons ensures eqns are satisfied
>>> cons([1,2,3])
[1, 2, 1.999999999999997]
>>> cons([5,5,1])
[5, 5.000000000000006, 1]
'''

# define an analytical function (i.e. non-symbolic)
def schedule(x0,x1,x2):
    return x0 * x1 - x2 * x2

# define the penalty condition
def penalty1(x): # <= 0.0
    x0,x1,x2 = x
    return 480 - schedule(x0,x1,x2)

# generate penalty from the penalty condition
@my.penalty.linear_inequality(penalty1)
def penalty(x):
    return 0.0

# generate constraint operator from the penalty
lb,ub = zip(*bounds)
c = my.constraints.as_constraint(penalty, lower_bounds=lb, upper_bounds=ub, nvars=3)

''' #NOTE: c ensures penalty1 is satisfied
>>> c([5,5,1])
[13.126545665004528, 44.97820356778645, 1.0138152329128338]
>>> schedule(*_)
589.3806217359323
>>> c([50,50,10])
[50.0, 50.0, 10.0]
>>> schedule(*_)
2400.0
'''

# define a constrait function for the input constraint
def intlog2(x):
    x[2] = np.round(np.log(x[0])**2)
    return x

# generate a constraint operator for all given constraints
ints = np.round
constraint = and_(c, cons, ints, intlog2)

''' #NOTE: constraint ensures c, cons, ints, and intlog2 are satisfied
>>> constraint([5,5,1])
[16.0, 42.0, 8.0]
>>> c(_)
[16.0, 42.0, 8.0]
>>> cons(_)
[16.0, 42.0, 8.0]
>>> intlog2(_)
[16.0, 42.0, 8.0]
>>> objective(_)
51050.0
'''

# solve
from mystic.solvers import diffev2
from mystic.monitors import VerboseMonitor
mon = VerboseMonitor(10)

result = diffev2(objective,x0=bounds, bounds=bounds, constraints=constraint, npop=20, gtol=50, disp=False, full_output=True, itermon=mon)

''' #NOTE: solution
>>> result[0]
array([ 14., 38.,   7.])
>>> result[1]
45450.0
'''
print(result[0])
print(result[1])



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2024-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

"""
First, minimize the Rosenbrock function f(x) under the following constraints:
  - mean(x) = 1
  - x_i are integers
  - x != [1, 1, 1]

Then, minimize the Rosenbrock function f(x) under the following constraints:
  - mean(x) = 2
  - x_i are unique
  - x_i are in the discrete set {0, 2, 4, 6, 8}
"""

from mystic.models import rosen
from mystic.solvers import diffev
from mystic.constraints import unique, discrete, integers, with_mean, and_, not_
from mystic.math.measures import mean

from mystic.bounds import Bounds
bounds = Bounds(0,10,n=3)

# generate the constraints
c = and_(integers()(lambda x:x), not_(lambda x:[1.]*len(x)), with_mean(1)(lambda x:x))

from mystic.monitors import VerboseMonitor
stepmon = VerboseMonitor(1)

result = diffev(rosen, x0=bounds, bounds=bounds, constraints=c, npop=100, gtol=10, disp=False, full_output=True, itermon=stepmon)

print("solved: %s @ %s" % (result[0], result[1]))
assert list(result[0]) == [0, 1, 2]
assert result[1] == 201

# generate the second set of constraints
c = and_(unique, discrete(range(0,10,2))(lambda x:x), with_mean(2)(lambda x:x))

stepmon = VerboseMonitor(1)

result = diffev(rosen, x0=bounds, bounds=bounds, constraints=c, npop=100, gtol=10, disp=False, full_output=True, itermon=stepmon)

print("solved: %s @ %s" % (result[0], result[1]))
assert list(result[0]) == [0, 2, 4]
assert result[1] == 402
